%{
#include "copyright.h"
#include "port.h"
#include "st.h"
#include "oct.h"
#include "utility.h"
#include "fang.h"
#include "harpoon.h"
#include "oh.h"
#include "tap.h"
#include "octmm.h"
#include "vov.h"

typedef struct layer_struct {
    char *name;
    fa_geometry *geos;
} layer_t;

typedef struct {
    fa_geometry *geos;
    int free_on_use;
} geo_t;
    
extern int yylineno;
extern FILE *yyin;
extern char yytext[];

static st_table *layer_table;
static octObject *facet;
static struct octBox facet_bb;
static char *filename;
static char *progname;
static int har_type = HAR_UNFILLED_BOX;

static double prescale;		/* Scale before running harpoon and fang. */
static double postscale;	/* Scale after running harpoon and fang. */
static double scale;		/* Cumulative scale factor. */
static int do_prescale;		/* Non-zero if a prescale factor was given. */
#define	PRE_SCALE \
	if (do_prescale) { \
		do_prescale = 0; \
		scale = prescale * postscale; \
		if (scale != 1.0) \
			OH_ASSERT(myScaleFacet(facet, scale)); \
	}

static enum st_retval fang_to_oct();
static void fetch_and_delete_objects();
static void delete_layer();
static layer_t *find_layer();
static void copy_labels();

extern octStatus flatten_contacts();
extern hackTerm();
extern void map_terminals_to_labels();
extern void flattenCell(), flattenSubcellLayer();
extern void squares();
%}

%union {
    int intval;
    double dblval;
    char *strval;
    geo_t layer;
}

%type <intval> integer
%type <dblval> real
%type <strval> name layer_name
%type <layer> expression
%left PLUS MINUS
%left TIMES XOR
%token NAME NUMBER REAL LPAREN RPAREN SEMI INVALID EQUAL COMMA SLASH
%token GROW SQUARES DELETE FRAME SCALE PROPERTY OPTIONS LABELS BOUNDBOX
%token GROWX GROWY
%token FRAMEX FRAMEY
%token FLATTEN SUBCELL
%start program

%%
program		: statement
		| program statement
		;

statement	: SEMI
 		    {}

		| layer_name EQUAL expression SEMI
		    {
			layer_t *dest = find_layer($1);
			if ($3.free_on_use) {
			    /* don't free $3, just re-assign it */
			    fa_free_geo(dest->geos);	/* discard lhs */
			    FREE(dest->geos);
			    dest->geos = $3.geos;
			} else {
			    /* need to copy geos -- watch for LAYER = LAYER; */
			    if ($3.geos != dest->geos) {
				fa_free_geo(dest->geos);      /* discard lhs */
				fa_copy($3.geos, dest->geos);
			    }
			}
			FREE($1);
		    }

		| OPTIONS options_list SEMI
		    {}
			
		| DELETE delete_list SEMI
		    {}

		| LABELS layer_name EQUAL layer_name SEMI
		    {
			copy_labels($4, $2);
		    }

		| PROPERTY name EQUAL name SEMI
		    {
			octObject prop;
			OH_ASSERT(ohCreateOrModifySprop(facet, &prop, $2, $4));
			FREE($2);
			FREE($4);
		    }

		| PROPERTY name EQUAL integer SEMI
		    {
			octObject prop;
			OH_ASSERT(ohCreateOrModifyIprop(facet, &prop, $2, $4));
			FREE($2);
		    }

		| PROPERTY name EQUAL real SEMI
		    {
			octObject prop;
			OH_ASSERT(ohCreateOrModifyRprop(facet, &prop, $2, $4));
			FREE($2);
		    }
		;


expression	: expression PLUS expression
		    {
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_or($1.geos, $3.geos, $$.geos);
			if ($1.free_on_use){fa_free_geo($1.geos);FREE($1.geos);}
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| expression TIMES expression
		    {
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_and($1.geos, $3.geos, $$.geos);
			if ($1.free_on_use){fa_free_geo($1.geos);FREE($1.geos);}
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| expression MINUS expression
		    {
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_andnot($1.geos, $3.geos, $$.geos);
			if ($1.free_on_use){fa_free_geo($1.geos);FREE($1.geos);}
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| expression XOR expression
		    {
			fa_geometry t1, t2;
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_andnot($1.geos, $3.geos, &t1);
			fa_andnot($3.geos, $1.geos, &t2);
			fa_or(&t1, &t2, $$.geos);
			if ($1.free_on_use){fa_free_geo($1.geos);FREE($1.geos);}
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
			fa_free_geo(&t1);
			fa_free_geo(&t2);
		    }

		| GROW LPAREN expression COMMA integer RPAREN 
		    {
			int i = $5 * postscale;
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_grow($3.geos, i, $$.geos);
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| GROWX LPAREN expression COMMA integer RPAREN 
		    {
			int i = $5 * postscale;
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_grow_xy($3.geos, i, FA_HORIZONTAL, $$.geos);
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| GROWY LPAREN expression COMMA integer RPAREN 
		    {
			int i = $5 * postscale;
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_grow_xy($3.geos, i, FA_VERTICAL, $$.geos);
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| SQUARES LPAREN expression COMMA integer COMMA integer
						COMMA integer RPAREN
		    {
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			squares($3.geos, $5, $7, $9, $$.geos);
			if ($3.free_on_use) {
			    fa_free_geo($3.geos);
			    FREE($3.geos);
			}
		    }

		| FRAME LPAREN expression COMMA integer RPAREN 
		    {
			int i = $5 * postscale;
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_frame($3.geos, i, i, $$.geos);
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| FRAMEX LPAREN expression COMMA integer RPAREN 
		    {
			int i = $5 * postscale;
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_frame_xy($3.geos, i, i, FA_HORIZONTAL, $$.geos);
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| FRAMEY LPAREN expression COMMA integer RPAREN 
		    {
			int i = $5 * postscale;
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_frame_xy($3.geos, i, i, FA_VERTICAL, $$.geos);
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| SCALE LPAREN expression COMMA real RPAREN 
		    {
			$$.geos = ALLOC(fa_geometry, 1);
			$$.free_on_use = 1;
			fa_scale($3.geos, $5, $$.geos);
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| LPAREN expression RPAREN
		    {
			$$ = $2;
		    }

		| BOUNDBOX LPAREN integer COMMA integer COMMA
                         integer COMMA integer RPAREN
		    {
			$$.geos = ALLOC(fa_geometry, 1);
			fa_init($$.geos);
			$$.free_on_use = 1;
			fa_add_box($$.geos, $3, $5, $7, $9);
		    }
			

		| BOUNDBOX LPAREN RPAREN
		    {
			$$.geos = ALLOC(fa_geometry, 1);
			fa_init($$.geos);
			$$.free_on_use = 1;
			fa_add_box($$.geos, 
			    facet_bb.lowerLeft.x, facet_bb.lowerLeft.y, 
			    facet_bb.upperRight.x, facet_bb.upperRight.y);
		    }
			
		| BOUNDBOX LPAREN expression RPAREN
		    {
			fa_box bb;
			$$.geos = ALLOC(fa_geometry, 1);
			fa_init($$.geos);
			$$.free_on_use = 1;
			fa_bb($3.geos, &bb);
			fa_add_box($$.geos, bb.left, bb.bottom, 
						bb.right, bb.top);
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
		    }

		| SUBCELL LPAREN expression RPAREN
		    {
			fa_geometry t1;
			$$.geos = ALLOC(fa_geometry, 1);
			flattenSubcellLayer(facet, $3.geos, &t1, $$.geos);
			$$.free_on_use = 1;
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
			fa_free_geo(&t1);
		    }

		| FLATTEN LPAREN expression RPAREN
		    {
			fa_geometry t1;
			$$.geos = ALLOC(fa_geometry, 1);
			flattenSubcellLayer(facet, $3.geos, $$.geos, &t1);
			$$.free_on_use = 1;
			if ($3.free_on_use){fa_free_geo($3.geos);FREE($3.geos);}
			fa_free_geo(&t1);
		    }

		| layer_name
		    {
			layer_t *layer;

			PRE_SCALE
			layer = find_layer($1);
			$$.geos = layer->geos;
			$$.free_on_use = 0;
			FREE($1);
		    }
		;


delete_list	: /* empty */
		| delete_list layer_name
		    {
			delete_layer($2);
			FREE($2);
		    }
	    	;

options_list	: /* empty */
		| options_list option
		;

option		: name
		    {
			if (strcmp($1, "flatten_connectors") == 0) {
			    (void) flatten_contacts(facet);
			} else if (strcmp($1, "infer_terminals") == 0) {
			    (void) hackTerm(facet, 1);
			} else if (strcmp($1, "map_terminals_to_labels") == 0) {
			    map_terminals_to_labels(progname, facet, scale);
			} else if (strcmp($1, "unfilled_polygons") == 0) {
			    har_type = HAR_UNFILLED_POLYGON;
			} else {
			    yyerror("unrecognized option");
			}
			FREE($1);
		    }

		| name EQUAL real
		    {
			if (strcmp($1, "prescale") == 0) {
			    prescale = prescale * $3;
			} else if (strcmp($1, "postscale") == 0) {
			    postscale = postscale * $3;
			} else {
			    yyerror("unrecognized option");
			}
			FREE($1);
		    }

		| SCALE EQUAL real
		    {
			prescale = prescale * $3;
		    }

		| name EQUAL integer SLASH integer
		    {
			double scl = (double) $3 / (double) $5;
			if (strcmp($1, "prescale") == 0) {
			    prescale = prescale * scl;
			} else if (strcmp($1, "postscale") == 0) {
			    postscale = postscale * scl;
			} else {
			    yyerror("unrecognized option");
			}
			FREE($1);
		    }

		| SCALE EQUAL integer SLASH integer
		    {
			double scl = (double) $3 / (double) $5;
			prescale = prescale * scl;
		    }
		;

layer_name	: NAME
		    { $$ = util_strsav(yytext); }
/* Layer names can't be a NUMBER because it confuses yacc. Thus, all
   layer names must start with a letter. This is an insignificant
   restriction. (PBC).

		| NUMBER
		    { $$ = util_strsav(yytext); }
*/
		;

name		: NAME
		    { $$ = util_strsav(yytext); }
		;

real		: REAL
		    { $$ = atof(yytext); }
		;

integer		: NUMBER
		    { $$ = atoi(yytext); }
		;
%%

/* utility routines - part of my not included in oh */

octStatus
myScaleContents(container, mask, scale)
octObject *container;
octObjectMask mask;
double scale;
{
    octObject geo;
    octGenerator gen;
    octStatus status;
    octObject inst;

    status = octInitGenContents(container, mask, &gen);
    if (status != OCT_OK) return status;

    while ((status = octGenerate(&gen, &geo)) == OCT_OK) {
	if (geo.type == OCT_INSTANCE) {
	    /* Do not scale the rotation matrix. (BCR) */
	    inst = geo;
	    octScaleGeo(&inst, scale);
	    geo.contents.instance.transform.translation =
			inst.contents.instance.transform.translation;
	    status = octModify(&geo);
	} else {
	    status = octScaleAndModifyGeo(&geo, scale);
	}
	if (status != OCT_OK) return status;
    }
    if (status != OCT_GEN_DONE) return status;
    return OCT_OK;
}


octStatus
myScaleFacet(facet, scale)
octObject *facet;
double scale;
{
    octObject layer;
    octGenerator gen;
    octStatus status;

    status = octInitGenContents(facet, OCT_LAYER_MASK, &gen);
    if (status != OCT_OK) return status;

    while ((status = octGenerate(&gen, &layer)) == OCT_OK) {
	status = myScaleContents(&layer, OCT_GEO_MASK, scale);
	if (status != OCT_OK) return status;
    }
    if (status != OCT_GEN_DONE) return status;

    status = myScaleContents(facet, OCT_INSTANCE_MASK, scale);
    if (status != OCT_GEN_DONE) return status;
    return OCT_OK;
}


/* start of octmm routines */

void
oct_mask_mod(the_progname, the_filename, the_facet, the_scale)
char *the_progname;
char *the_filename;
octObject *the_facet;
char *the_scale;
{
    octStatus status;
    char temp1[1024];
    double a, b;
    char *the_post;

    /* Initialize layer identifiers for flattening. */
    flattenCell();

    /* Set the (ack!) global variables */
    layer_table = st_init_table(strcmp, st_strhash);
    facet = the_facet;
    progname = (the_progname == NIL(char)) ? "" : the_progname;
    scale = 1.0;
    prescale = 1.0;
    postscale = 1.0;
    do_prescale = 1;

    /* Determine the command-line pre-scale */
    if (the_scale != NIL(char)) {
	if ((the_post = strchr(the_scale, 'x')) != NIL(char)) {
	    *the_post++ = '\000';
	    if (*the_post != '\000') {
		if (strchr(the_post, '/') == NIL(char)) {
		    postscale = atof(the_post);
		} else  {
		    (void) sscanf(the_post, "%lf/%lf", &a, &b);
		    postscale = a / b;
		}
	    }
	}
	if (*the_scale != '\000') {
	    if (strchr(the_scale, '/') == NIL(char)) {
		prescale = atof(the_scale);
	    } else {
		(void) sscanf(the_scale, "%lf/%lf", &a, &b);
		prescale = a / b;
	    }
	}
	if (the_post) *--the_post = 'x';
    }

    /* Open the program file */
    filename = the_filename;
    if (filename == NIL(char) || strcmp(filename, "") == 0) {
	/* There is no OCTMM program, so only scale the facet. */
	PRE_SCALE
	return;
    }

    if ((yyin = VOVfopen(the_filename, "r")) == NULL) {
	(void) sprintf(temp1, "%s/%s",
		       util_file_expand(tapGetDefaultRoot()),
		       the_filename);
	filename = temp1;
	if ((yyin = VOVfopen(filename, "r")) == NULL) {
	    perror(filename);
	    exit(1);
	}
    }

    /* Find BBOX of the facet */
    status = octBB(facet, &facet_bb);
    if (status == OCT_NO_BB) {
	facet_bb.lowerLeft.x = 0;
	facet_bb.lowerLeft.y = 0;
	facet_bb.upperRight.x = 0;
	facet_bb.upperRight.y = 0;
    } else {
	OH_ASSERT(status);
    }

    /* Use our friend yacc to parse things */
    if (yyparse() != 0) {
	(void) fprintf(stderr, "%s: error in parse\n", progname);
	exit(1);
    }

    /* Place all remaining layers back into the facet */
    (void) st_foreach(layer_table, fang_to_oct, NIL(char));
    st_free_table(layer_table);
    (void) fclose(yyin);
}


/* ARGSUSED */
static enum st_retval
fang_to_oct(key, value, arg)
char *key;
char *value;
char *arg;
{
    layer_t *layer = (layer_t *) value;
    octObject oct_layer;

    OH_ASSERT(ohGetOrCreateLayer(facet, &oct_layer, layer->name));
    harFangToContents(layer->geos, &oct_layer, har_type);
    fa_free_geo(layer->geos);
    FREE(layer->geos);
    FREE(layer->name);
    FREE(layer); 
    return ST_CONTINUE;
}


static layer_t *
find_layer(name)
char *name;
{
    octObject oct_layer;
    octStatus status;
    layer_t *layer;
    char *layer_p;

    if (st_lookup(layer_table, name, &layer_p) == 0) {
        status = ohGetByLayerName(facet, &oct_layer, name);
	if (status == OCT_OK || status == OCT_NOT_FOUND) {
	    layer = ALLOC(layer_t, 1);
	    layer->name = util_strsav(name);
	    layer->geos = ALLOC(fa_geometry, 1);
	    fa_init(layer->geos);
	    (void) st_insert(layer_table, layer->name, (char *) layer);
	    if (status == OCT_OK) {
		fetch_and_delete_objects(&oct_layer, layer->geos);
	    }
	} else {
	    OH_ASSERT(status);
	}
    } else {
	layer = (layer_t *) layer_p;
    }
    return layer;
}


/*
 *  Move objects from octObject 'layer' to a fa_geometry 'geos', and then
 *  delete the objects from Oct.  Do NOT delete labels or terminal
 *  implementations.
 */
static void
fetch_and_delete_objects(layer, geos)
octObject *layer;
fa_geometry *geos;
{
    octObject fterm, obj;
    octGenerator gen;
    octStatus status;

    OH_ASSERT(octInitGenContents(layer, OCT_GEO_MASK &~ OCT_LABEL_MASK, &gen));
    while ((status = octGenerate(&gen, &obj)) == OCT_OK) {
	harAddObjectToFang(&obj, geos, HAR_NO_TRANSFORM);

	/* Is this a terminal implementation ?? */
	status = octGenFirstContainer(&obj, OCT_TERM_MASK, &fterm);
	if (status == OCT_GEN_DONE) {
	    OH_ASSERT(octDelete(&obj));	/* not a terminal frame */
	} else {
	    OH_ASSERT(status);			/* make sure status was okay */
	}
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
}


static void
delete_layer( name)
char *name;
{
    char *layer_p, *dummy;
    layer_t *layer;
    octObject oct_layer, obj;
    octGenerator gen;
    octStatus status;

    if (st_lookup(layer_table, name, &layer_p) != 0) {
	/* Throw away the existing geo */
	layer = (layer_t *) layer_p;
	dummy = name;
	(void) st_delete(layer_table, &dummy, NIL(char *));
	fa_free_geo(layer->geos);
	FREE(layer->geos);
	FREE(layer->name);
	FREE(layer);
    }

    /* Delete the oct objects on the layer, and the layer */
    if (ohGetByLayerName(facet, &oct_layer, name) == OCT_OK) {
	OH_ASSERT(octInitGenContents(&oct_layer, OCT_ALL_MASK, &gen));
	while ((status = octGenerate(&gen, &obj)) == OCT_OK) {
	    OH_ASSERT(octDelete(&obj));
	}
	OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
	OH_ASSERT(octDelete(&oct_layer));
    }
}


static void
copy_labels(src_layer, dst_layer)
char *src_layer, *dst_layer;
{
    octObject lyr, slayer, dlayer, label, fterm, geo;
    octGenerator gen, gen1;
    octStatus status;

    /* Get the source and destination layers */
    OH_ASSERT(ohGetOrCreateLayer(facet, &slayer, src_layer));
    OH_ASSERT(ohGetOrCreateLayer(facet, &dlayer, dst_layer));

    /* Move labels from one layer to the other */
    OH_ASSERT(octInitGenContents(&slayer, OCT_LABEL_MASK, &gen));
    while ((status = octGenerate(&gen, &label)) == OCT_OK) {
	OH_ASSERT(octDetach(&slayer, &label));
	OH_ASSERT(octAttach(&dlayer, &label));
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);

    /* Copy terminal implementation attachments (needed by ciftooct) */
    OH_ASSERT(octInitGenContents(facet, OCT_TERM_MASK, &gen));
    while ((status = octGenerate(&gen, &fterm)) == OCT_OK) {
	if (!octIdIsNull(fterm.contents.term.instanceId)) continue;

	/* Find all geos attached to this terminal */
	OH_ASSERT(octInitGenContents(&fterm, OCT_GEO_MASK, &gen1));
	while ((status = octGenerate(&gen1, &geo)) == OCT_OK) {

	    /* find what layer this geo is on */
	    status = octGenFirstContainer(&geo, OCT_LAYER_MASK, &lyr);
	    if (status == OCT_OK) {
		if (octIdsEqual(lyr.objectId, slayer.objectId)) {
		    OH_ASSERT(octCreate(&dlayer, &geo));
		    OH_ASSERT(octAttach(&fterm, &geo));
		}
	    } else if (status != OCT_GEN_DONE) {
		OH_ASSERT(status);
	    }
	}
	OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
}


int
yyerror(s)
char *s;
{
    (void) fprintf(stderr, "%s: %s in file %s at line %d\n", 
	progname, s, filename, yylineno);
    exit(1);
}


int
yywrap()
{
    return 1;
}

