/*
 * ------------------------------------------------------------------------
 *      PACKAGE:  [incr Tk]
 *  DESCRIPTION:  Building mega-widgets with [incr Tcl]
 *
 *  [incr Tk] provides a framework for building composite "mega-widgets"
 *  using [incr Tcl] classes.  It defines a set of base classes that are
 *  specialized to create all other widgets.
 *
 *  This part adds C implementations for some of the methods in the
 *  base class itk::Archetype.
 *
 *    Itk_ArchComponentCmd   <=> itk_component
 *    Itk_ArchOptionCmd      <=> itk_option
 *    Itk_ArchInitCmd        <=> itk_initialize
 *    Itk_ArchCompAccessCmd  <=> component
 *    Itk_ArchConfigureCmd   <=> configure
 *    Itk_ArchCgetCmd        <=> cget
 *
 *    Itk_ArchInitOptsCmd    <=> _initOptionInfo (used to set things up)
 *    Itk_ArchDeleteOptsCmd  <=> _deleteOptionInfo (used to clean things up)
 *
 * ========================================================================
 *  AUTHOR:  Michael J. McLennan       Phone: (610)712-2842
 *           AT&T Bell Laboratories   E-mail: michael.mclennan@att.com
 *     RCS:  $Id: itk_archetype.c,v 1.4 1995/10/25 05:10:55 mmc Exp $
 * ========================================================================
 *             Copyright (c) 1993-1995  AT&T Bell Laboratories
 * ------------------------------------------------------------------------
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
#include <assert.h>
#include "itkInt.h"

/*
 *  Info associated with each Archetype mega-widget:
 */
typedef struct ArchInfo {
    Itcl_Object *obj;           /* object containing this info */
    Tk_Window tkwin;            /* window representing this mega-widget */
    Tcl_HashTable components;   /* list of all mega-widget components */
    Tcl_HashTable options;      /* list of all mega-widget options */
    ItkOptList order;           /* gives ordering of options */
} ArchInfo;

/*
 *  Each component widget in an Archetype mega-widget:
 */
typedef struct ArchComponent {
    int protection;               /* protection level for this component */
    Tcl_Command accessCmd;        /* access command for component widget */
    Tk_Window tkwin;              /* Tk window for this component widget */
} ArchComponent;

/*
 *  Each option in an Archetype mega-widget:
 */
typedef struct ArchOption {
    char *switchName;             /* command-line switch for this option */
    char *resName;                /* resource name in X11 database */
    char *resClass;               /* resource class name in X11 database */
    char *init;                   /* initial value for option */
    int flags;                    /* flags representing option state */
    Itcl_List parts;              /* parts relating to this option */
} ArchOption;

/*
 *  Flag bits for ArchOption state:
 */
#define ITK_ARCHOPT_INIT  0x01    /* option has been initialized */

/*
 *  Various parts of a composite option in an Archetype mega-widget:
 */
typedef int (Itk_ConfigOptionPartProc) _ANSI_ARGS_((Tcl_Interp *interp,
    Itcl_Object *obj, ClientData cdata, char* newVal));

typedef struct ArchOptionPart {
    ClientData clientData;                 /* data associated with this part */
    Itk_ConfigOptionPartProc *configProc;  /* update when new vals arrive */
    Itcl_DeleteProc *deleteProc;           /* clean up after clientData */

    ClientData from;                       /* token that indicates who
                                            * contributed this option part */
} ArchOptionPart;


/*
 *  Info kept by the itk::option-parser namespace and shared by
 *  all option processing commands:
 */
typedef struct ArchMergeInfo {
    Tcl_HashTable usualCode;      /* usual option handling code for the
                                   * various widget classes */

    ArchInfo *archInfo;           /* internal option info for mega-widget */
    ArchComponent *archComp;      /* component being merged into mega-widget */
    Tcl_HashTable *optionTable;   /* table of valid configuration options
                                   * for component being merged */
} ArchMergeInfo;

/*
 *  Used to capture component widget configuration options when a
 *  new component is being merged into a mega-widget:
 */
typedef struct GenericConfigOpt {
    char *switchName;             /* command-line switch for this option */
    char *resName;                /* resource name in X11 database */
    char *resClass;               /* resource class name in X11 database */
    char *init;                   /* initial value for this option */
    char *value;                  /* current value for this option */
    char **storage;               /* storage for above strings */

    ArchOption *integrated;       /* integrated into this mega-widget option */
    ArchOptionPart *optPart;      /* integrated as this option part */
} GenericConfigOpt;


/* list of all objects with extra Archetype info */
static Tcl_HashTable* ObjsWithArchInfo = NULL;


#ifndef lint
static char rcsid[] = "$Id: itk_archetype.c,v 1.4 1995/10/25 05:10:55 mmc Exp $";
#endif

/*
 *  FORWARD DECLARATIONS
 */
static void Itk_DelMergeInfo _ANSI_ARGS_((ClientData cdata));

static int Itk_ArchInitOptsCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static void Itk_DelArchInfo _ANSI_ARGS_((ClientData cdata));
static int Itk_ArchDeleteOptsCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));

static int Itk_ArchComponentCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchCompAddCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchCompDeleteCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchOptKeepCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchOptIgnoreCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchOptRenameCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchOptUsualCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));

static int Itk_ArchInitCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchOptionCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchOptionAddCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchOptionRemoveCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));

static int Itk_ArchCompAccessCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchConfigureCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_ArchCgetCmd _ANSI_ARGS_((ClientData cdata,
    Tcl_Interp *interp, int argc, char **argv));
static int Itk_PropagateOption _ANSI_ARGS_((Tcl_Interp *interp,
    Itcl_Object *obj, ClientData cdata, char *newval));

static int Itk_ArchSetOption _ANSI_ARGS_((Tcl_Interp *interp,
    ArchInfo *info, char *name, char *value));
static int Itk_ArchConfigOption _ANSI_ARGS_((Tcl_Interp *interp,
    ArchInfo *info, char *name, char *value, int uplevel));
static void Itk_ArchOptConfigError _ANSI_ARGS_((Tcl_Interp *interp,
    ArchInfo *info, ArchOption *archOpt));
static void Itk_ArchOptAccessError _ANSI_ARGS_((Tcl_Interp *interp,
    ArchInfo *info, ArchOption *archOpt));

static int Itk_GetArchInfo _ANSI_ARGS_((Tcl_Interp *interp,
    Itcl_Object* obj, ArchInfo **infoPtr));

static ArchComponent* Itk_CreateArchComponent _ANSI_ARGS_((
    Tcl_Interp *interp, ArchInfo *info, char *name,
    Tcl_Command accessCmd));
static void Itk_DelArchComponent _ANSI_ARGS_((ArchComponent *archComp));

static int Itk_GetArchOption _ANSI_ARGS_((Tcl_Interp *interp,
    ArchInfo *info, char *switchName, char *resName, char *resClass,
    char *defVal, char *currVal, ArchOption **aoPtr));
static void Itk_DelArchOption _ANSI_ARGS_((ArchOption *archOpt));

static ArchOptionPart* Itk_CreateOptionPart _ANSI_ARGS_((
    Tcl_Interp *interp, ClientData cdata, Itk_ConfigOptionPartProc* cproc,
    Itcl_DeleteProc *dproc, ClientData from));
static int Itk_AddOptionPart _ANSI_ARGS_((Tcl_Interp *interp,
    ArchInfo *info, char *switchName, char *resName, char *resClass,
    char *defVal, char *currVal, ArchOptionPart *optPart,
    ArchOption **raOpt));
static ArchOptionPart* Itk_FindArchOptionPart _ANSI_ARGS_((
    ArchInfo *info, char *switchName, ClientData from));
static int Itk_RemoveArchOptionPart _ANSI_ARGS_((ArchInfo *info,
    char *switchName, ClientData from));
static int Itk_IgnoreArchOptionPart _ANSI_ARGS_((ArchInfo *info,
    GenericConfigOpt *opt));
static void Itk_DelOptionPart _ANSI_ARGS_((ArchOptionPart *optPart));

static Tcl_HashTable* Itk_CreateGenericOptTable _ANSI_ARGS_((Tcl_Interp *interp,
    char *options));
static void Itk_DelGenericOptTable _ANSI_ARGS_((Tcl_HashTable *tPtr));

static GenericConfigOpt* Itk_CreateGenericOpt _ANSI_ARGS_((Tcl_Interp *interp,
    char *switchName, Tcl_Command accessCmd));
static void Itk_DelGenericOpt _ANSI_ARGS_((GenericConfigOpt* opt));


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchetypeInit()
 *
 *  Invoked by Itk_Init() whenever a new interpreter is created to
 *  declare the procedures used in the itk::Archetype base class.
 * ------------------------------------------------------------------------
 */
int
Itk_ArchetypeInit(interp)
    Tcl_Interp *interp;  /* interpreter to be updated */
{
    ArchMergeInfo *mergeInfo;
    Itcl_Namespace parserNs;

    /*
     *  Declare all of the C routines that are integrated into
     *  the Archetype base class.
     */
    if (Itcl_RegisterC(interp,
            "Archetype-init", Itk_ArchInitOptsCmd) != TCL_OK ||

        Itcl_RegisterC(interp,
            "Archetype-delete", Itk_ArchDeleteOptsCmd) != TCL_OK ||

        Itcl_RegisterC(interp,
            "Archetype-itk_component", Itk_ArchComponentCmd) != TCL_OK ||

        Itcl_RegisterC(interp,
            "Archetype-itk_option", Itk_ArchOptionCmd) != TCL_OK ||

        Itcl_RegisterC(interp,
            "Archetype-itk_initialize", Itk_ArchInitCmd) != TCL_OK ||

        Itcl_RegisterC(interp,
            "Archetype-component", Itk_ArchCompAccessCmd) != TCL_OK ||

        Itcl_RegisterC(interp,
            "Archetype-configure",Itk_ArchConfigureCmd) != TCL_OK ||

        Itcl_RegisterC(interp,
            "Archetype-cget",Itk_ArchCgetCmd) != TCL_OK) {

        return TCL_ERROR;
    }

    /*
     *  Create the namespace containing the option parser commands.
     */
    mergeInfo = (ArchMergeInfo*)ckalloc(sizeof(ArchMergeInfo));
    Tcl_InitHashTable(&mergeInfo->usualCode,TCL_STRING_KEYS);
    mergeInfo->archInfo    = NULL;
    mergeInfo->archComp    = NULL;
    mergeInfo->optionTable = NULL;

    if (Itcl_CreateNamesp(interp, "::itk::option-parser",
        Itcl_PreserveData((ClientData)mergeInfo), Itcl_ReleaseData,
        &parserNs) != TCL_OK) {

        Itk_DelMergeInfo((ClientData)mergeInfo);
        Tcl_AppendResult(interp, " (cannot initialize itk option parser)",
            (char*)NULL);
        return TCL_ERROR;
    }
    Itcl_EventuallyFree((ClientData)mergeInfo, Itk_DelMergeInfo);

    Tcl_CreateCommand(interp, "::itk::option-parser::keep",
        Itk_ArchOptKeepCmd,
        (ClientData)mergeInfo, (Tcl_CmdDeleteProc*)NULL);

    Tcl_CreateCommand(interp, "::itk::option-parser::ignore",
        Itk_ArchOptIgnoreCmd,
        (ClientData)mergeInfo, (Tcl_CmdDeleteProc*)NULL);

    Tcl_CreateCommand(interp, "::itk::option-parser::rename",
        Itk_ArchOptRenameCmd,
        (ClientData)mergeInfo, (Tcl_CmdDeleteProc*)NULL);

    Tcl_CreateCommand(interp, "::itk::option-parser::usual",
        Itk_ArchOptUsualCmd,
        (ClientData)mergeInfo, (Tcl_CmdDeleteProc*)NULL);

    /*
     *  Add the "itk::usual" command to register option handling code.
     */
    Tcl_CreateCommand(interp, "::itk::usual", Itk_UsualCmd,
        Itcl_PreserveData((ClientData)mergeInfo), Itcl_ReleaseData);

    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_DelMergeInfo()
 *
 *  Destroys the "merge" info record shared by commands in the
 *  itk::option-parser namespace.  Invoked automatically when the
 *  namespace containing the parsing commands is destroyed.
 * ------------------------------------------------------------------------
 */
static void
Itk_DelMergeInfo(cdata)
    ClientData cdata;  /* data to be destroyed */
{
    ArchMergeInfo *mergeInfo = (ArchMergeInfo*)cdata;

    char *code;
    Tcl_HashEntry *entry;
    Tcl_HashSearch place;

    assert(mergeInfo->optionTable == NULL);

    entry = Tcl_FirstHashEntry(&mergeInfo->usualCode, &place);
    while (entry) {
        code = (char*)Tcl_GetHashValue(entry);
        ckfree(code);
        entry = Tcl_NextHashEntry(&place);
    }
    Tcl_DeleteHashTable(&mergeInfo->usualCode);

    ckfree((char*)mergeInfo);
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchInitOptsCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::_initOptionInfo
 *  method.  This method should be called out in the constructor for
 *  each object, to initialize the object so that it can be used with
 *  the other access methods in this file.  Allocates some extra
 *  data associated with the object at the C-language level.
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchInitOptsCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    ArchInfo *info;
    Itcl_Namespace classNs;
    Itcl_Class *cdefn;
    Itcl_Object *obj;
    Tcl_HashEntry *entry;
    int newEntry;

    if (argc != 1) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
            argv[0], "\"",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (Itcl_GetClassContext(interp,&classNs,&cdefn,&obj) != TCL_OK || !obj) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "cannot use \"", argv[0],
            "\" without an object context",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Create the info to track all Archetype objects.
     */
    if (ObjsWithArchInfo == NULL) {
        ObjsWithArchInfo = (Tcl_HashTable*)ckalloc(sizeof(Tcl_HashTable));
        Tcl_InitHashTable(ObjsWithArchInfo, TCL_ONE_WORD_KEYS);
    }

    /*
     *  Create some info for the current object and register
     *  it on the list of all objects.
     */
    info = (ArchInfo*)ckalloc(sizeof(ArchInfo));
    info->obj = obj;
    info->tkwin = NULL;  /* not known yet */
    Tcl_InitHashTable(&info->components,TCL_STRING_KEYS);
    Tcl_InitHashTable(&info->options,TCL_STRING_KEYS);
    Itk_OptListInit(&info->order, &info->options);

    entry = Tcl_CreateHashEntry(ObjsWithArchInfo, (char*)obj, &newEntry);
    if (!newEntry) {
        Itk_DelArchInfo( (ArchInfo*)Tcl_GetHashValue(entry) );
    }
    Tcl_SetHashValue(entry, (ClientData)info);

    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_DelArchInfo()
 *
 *  Invoked when the option info associated with an itk::Archetype
 *  widget is no longer needed.  This usually happens when a widget
 *  is destroyed.  Frees the given bundle of data and removes it
 *  from the global list of Archetype objects.
 * ------------------------------------------------------------------------
 */
static void
Itk_DelArchInfo(cdata)
    ClientData cdata;    /* client data for Archetype objects */
{
    ArchInfo *info = (ArchInfo*)cdata;

    Tcl_HashEntry *entry;
    Tcl_HashSearch place;
    ArchOption *archOpt;
    ArchComponent *archComp;

    /*
     *  Destroy all component widgets.
     */
    entry = Tcl_FirstHashEntry(&info->components,&place);
    while (entry) {
        archComp = (ArchComponent*)Tcl_GetHashValue(entry);
        Itk_DelArchComponent(archComp);
        entry = Tcl_NextHashEntry(&place);
    }
    Tcl_DeleteHashTable(&info->components);

    /*
     *  Destroy all information associated with configuration options.
     */
    entry = Tcl_FirstHashEntry(&info->options,&place);
    while (entry) {
        archOpt = (ArchOption*)Tcl_GetHashValue(entry);
        Itk_DelArchOption(archOpt);
        entry = Tcl_NextHashEntry(&place);
    }
    Tcl_DeleteHashTable(&info->options);
    Itk_OptListFree(&info->order);

    ckfree((char*)info);
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchDeleteOptsCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::_deleteOptionInfo
 *  method.  This method should be called out in the destructor for each
 *  object, to clean up data allocated by Itk_ArchInitOptsCmd().
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchDeleteOptsCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    Tcl_HashEntry *entry = NULL;

    Itcl_Namespace classNs;
    Itcl_Class *cdefn;
    Itcl_Object *obj;

    if (argc != 1) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
            argv[0], "\"",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (Itcl_GetClassContext(interp,&classNs,&cdefn,&obj) != TCL_OK || !obj) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "cannot use \"", argv[0],
            "\" without an object context",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Find the info associated with this object.
     *  Destroy the data and remove it from the global list.
     */
    if (ObjsWithArchInfo != NULL) {
        entry = Tcl_FindHashEntry(ObjsWithArchInfo, (char*)obj);
    }
    if (entry) {
        Itk_DelArchInfo( Tcl_GetHashValue(entry) );
        Tcl_DeleteHashEntry(entry);
    }
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchComponentCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::itk_component
 *  method.  Handles the following options:
 *
 *      itk_component add <name> <createCmds> ?<optionCmds>?
 *      itk_component delete <name> ?<name>...?
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchComponentCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    char c;
    int length;

    /*
     *  Check arguments and handle the various options...
     */
    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be one of...\n",
            "  ", argv[0], " add name createCmds ?optionCmds?\n",
            "  ", argv[0], " delete name ?name name...?",
            (char*)NULL);
        return TCL_ERROR;
    }
    c = *argv[1];
    length = strlen(argv[1]);

    /*
     *  Handle:  itk_component add...
     */
    if (c == 'a' && strncmp(argv[1], "add", length) == 0) {
        if (argc < 4 || argc > 5) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                argv[0], " add name createCmds ?optionCmds?\"",
                (char*)NULL);
            return TCL_ERROR;
        }
        return Itk_ArchCompAddCmd(dummy, interp, argc-1, argv+1);
    }

    /*
     *  Handle:  itk_component delete...
     */
    else if (c == 'd' && strncmp(argv[1], "delete", length) == 0) {
        if (argc < 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                argv[0], " delete name ?name name...?\"",
                (char*)NULL);
            return TCL_ERROR;
        }
        return Itk_ArchCompDeleteCmd(dummy, interp, argc-1, argv+1);
    }

    /*
     *  Flag any errors.
     */
    Tcl_AppendResult(interp, "bad option \"", argv[1],
        "\": should be one of...\n",
        "  ", argv[0], " add name createCmds ?optionCmds?\n",
        "  ", argv[0], " delete name ?name name...?",
        (char*)NULL);
    return TCL_ERROR;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchCompAddCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::itk_component
 *  method.  Adds a new component widget into the mega-widget,
 *  integrating its configuration options into the master list.
 *
 *      itk_component add ?-global? <name> <createCmds> <optionCmds>
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchCompAddCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    Tcl_HashEntry *entry = NULL;
    char *path = NULL;
    ArchComponent *archComp = NULL;
    ArchMergeInfo *mergeInfo = NULL;
    Tcl_DString buffer, objName;

    int newEntry, status;
    Itcl_Namespace classNs, parserNs, ns;
    Itcl_ActiveNamespace nsToken;
    Itcl_Class *cdefn;
    Itcl_Object *obj;
    ArchInfo *info;
    Tcl_Command accessCmd;

    Tcl_DStringInit(&buffer);
    Tcl_DStringInit(&objName);

    /*
     *  Get the Archetype info associated with this widget.
     */
    if (Itcl_GetClassContext(interp,&classNs,&cdefn,&obj) != TCL_OK || !obj) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "cannot access components without ",
            "an object context",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (Itk_GetArchInfo(interp,obj,&info) != TCL_OK) {
        return TCL_ERROR;
    }

    /*
     *  See if a component already exists with the symbolic name.
     */
    entry = Tcl_CreateHashEntry(&info->components, argv[1], &newEntry);
    if (!newEntry) {
        Tcl_AppendResult(interp, "component \"", argv[1],
            "\" already defined",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  If this component is the "hull" for the mega-widget, then
     *  move the object access command out of the way before
     *  creating the component, so it is not accidentally deleted.
     */
    if (strcmp(argv[1],"hull") == 0) {
        Tcl_DStringTrunc(&buffer, 0);
        Tcl_DStringAppendElement(&buffer, "rename");
        Tcl_DStringAppendElement(&buffer,
            Tcl_GetCommandPath(interp, obj->accessCmd));
        Tcl_DStringAppendElement(&buffer,
            Tcl_GetCommandPath(interp, obj->accessCmd));
        Tcl_DStringAppend(&buffer, "-widget-", -1);

        Tcl_DStringTrunc(&objName, 0);
        Tcl_DStringAppendElement(&objName, "rename");
        Tcl_DStringAppendElement(&objName,
            Tcl_GetCommandPath(interp, obj->accessCmd));
        Tcl_DStringAppend(&objName, "-widget-", -1);
        Tcl_DStringAppendElement(&objName,
            Tcl_GetCommandPath(interp, obj->accessCmd));

        if (Tcl_Eval(interp, Tcl_DStringValue(&buffer)) != TCL_OK) {
            goto compFail;
        }
    }

    /*
     *  Execute the <createCmds> to create the component widget.
     *  Do this one level up, in the scope of the calling routine.
     */
    if (Itcl_UplevelEval(interp, argv[2]) != TCL_OK) {
        goto compFail;
    }

    /*
     *  Take the result from the widget creation commands as the
     *  path name for the new component.  Make a local copy of
     *  this, since the interpreter will get used in the mean time.
     */
    path = (char*)ckalloc((unsigned)(strlen(interp->result)+1));
    strcpy(path, interp->result);

    /*
     *  Look for the access command token in the context of the
     *  calling namespace.  By-pass any protection at this point.
     */
    accessCmd = NULL;

    ns = Itcl_GetCallingNamesp(interp,1);
    nsToken = Itcl_ActivateNamesp(interp, ns);
    if (nsToken) {
        if (Itcl_FindCommand(interp, path, ITCL_FIND_NO_PRIVACY,
            &accessCmd) != TCL_OK) {
            goto compFail;
        }
        Itcl_DeactivateNamesp(interp, nsToken);
    }
    if (!accessCmd) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "cannot find component access command \"",
            path, "\" for component \"", argv[1], "\"",
            (char*)NULL);
        goto compFail;
    }

    /*
     *  Create the component record.
     */
    archComp = Itk_CreateArchComponent(interp, info, argv[1], accessCmd);

    if (!archComp) {
        goto compFail;
    }
    Tcl_SetHashValue(entry, (ClientData)archComp);

    /*
     *  If this component is the "hull" for the mega-widget, then
     *  move the hull widget access command to a different name,
     *  and move the object access command back into place.  This
     *  way, when the widget name is used as a command, the object
     *  access command will handle all requests.
     */
    if (strcmp(argv[1],"hull") == 0) {
        Tcl_DStringTrunc(&buffer, 0);
        Tcl_DStringAppendElement(&buffer, "rename");
        Tcl_DStringAppendElement(&buffer, 
            Tcl_GetCommandPath(interp, accessCmd));
        Tcl_DStringAppendElement(&buffer,
            Tcl_GetCommandPath(interp, accessCmd));
        Tcl_DStringAppend(&buffer, "-itk_hull", -1);

        if (Tcl_Eval(interp, Tcl_DStringValue(&buffer)) != TCL_OK) {
            goto compFail;
        }
        if (Tcl_Eval(interp, Tcl_DStringValue(&objName)) != TCL_OK) {
            goto compFail;
        }
    }

    /*
     *  Register the mega-widget container for this component with Tk.
     *  Also, add a binding onto the new component, so that when its
     *  window is destroyed, it will automatically remove itself
     *  from its parent's component list.  Avoid doing these things
     *  for the "hull" component, since it is a special case and
     *  these things are not really necessary.
     */
    if (strcmp(argv[1],"hull") != 0) {
        if (archComp->tkwin && info->tkwin) {
            Itk_SetMegaWidget(archComp->tkwin, info->tkwin);
        }

        Tcl_DStringTrunc(&buffer, 0);
        Tcl_DStringAppend(&buffer, "bindtags ", -1);
        Tcl_DStringAppend(&buffer, path, -1);
        if (Tcl_Eval(interp, Tcl_DStringValue(&buffer)) != TCL_OK) {
            goto compFail;
        }

        Tcl_DStringTrunc(&buffer, 0);
        Tcl_DStringAppend(&buffer, "bind itk-destroy-", -1);
        Tcl_DStringAppend(&buffer, path, -1);
        Tcl_DStringAppend(&buffer, " <Destroy> [code ", -1);
        Tcl_DStringAppend(&buffer,
            Tcl_GetCommandPath(interp, obj->accessCmd), -1);
        Tcl_DStringAppend(&buffer, " itk_component delete ", -1);
        Tcl_DStringAppend(&buffer, argv[1], -1);
        Tcl_DStringAppend(&buffer, "]\n", -1);
        Tcl_DStringAppend(&buffer, "bindtags ", -1);
        Tcl_DStringAppend(&buffer, path, -1);
        Tcl_DStringAppend(&buffer, " {itk-destroy-", -1);
        Tcl_DStringAppend(&buffer, path, -1);
        Tcl_DStringAppend(&buffer, " ", -1);
        Tcl_DStringAppend(&buffer, interp->result, -1);
        Tcl_DStringAppend(&buffer, "}", -1);
        if (Tcl_Eval(interp, Tcl_DStringValue(&buffer)) != TCL_OK) {
            goto compFail;
        }
    }

    /*
     *  Query the list of configuration options for this widget,
     *  so we will know which ones are valid.  Build an option
     *  table to represent these, so they can be found quickly
     *  by the option parsing commands in "itk::option-parser".
     */
    Tcl_DStringTrunc(&buffer, 0);
    Tcl_DStringAppendElement(&buffer, Tcl_GetCommandPath(interp, accessCmd));
    Tcl_DStringAppendElement(&buffer, "configure");

    status  = Itcl_NamespEval(interp,
        Itcl_GetCommandNamesp(archComp->accessCmd),
        Tcl_DStringValue(&buffer));

    if (status != TCL_OK) {
        goto compFail;
    }
    Tcl_DStringTrunc(&buffer, 0);
    Tcl_DStringAppend(&buffer, interp->result, -1);

    /*
     *  Find the "itk::option-parser" namespace and get the data
     *  record shared by all of the parsing commands.
     */
    if (Itcl_FindNamesp(interp, "::itk::option-parser", 0, &parserNs)
        != TCL_OK) {
        goto compFail;
    }
    if (parserNs == NULL) {
        Tcl_AppendResult(interp, "internal error: ",
            "cannot find ::itk::option-parser namespace",
            (char*)NULL);
        goto compFail;
    }
    mergeInfo = (ArchMergeInfo*)Itcl_GetNamespData(parserNs);
    assert(mergeInfo);

    /*
     *  Initialize the data record used by the option parsing commands.
     *  Store a table of valid configuration options, along with the
     *  info for the mega-widget that is being updated.
     */
    mergeInfo->optionTable = Itk_CreateGenericOptTable(interp,
        Tcl_DStringValue(&buffer));

    if (!mergeInfo->optionTable) {
        goto compFail;
    }
    mergeInfo->archInfo = info;
    mergeInfo->archComp = archComp;

    /*
     *  Execute the option-handling commands in the "itk::option-parser"
     *  namespace.  If there are no option-handling commands, invoke
     *  the "usual" command instead.
     */
    if (argc == 4) {
        status = Itcl_NamespEval(interp, parserNs, argv[3]);
    } else {
        status = Itcl_NamespEval(interp, parserNs, "usual");
    }
    if (status != TCL_OK) {
        goto compFail;
    }

    Itk_DelGenericOptTable(mergeInfo->optionTable);
    mergeInfo->optionTable = NULL;
    mergeInfo->archInfo    = NULL;
    mergeInfo->archComp    = NULL;

    ckfree(path);

    Tcl_DStringFree(&buffer);
    Tcl_DStringFree(&objName);

    Tcl_SetResult(interp, argv[1], TCL_VOLATILE);
    return TCL_OK;

    /*
     *  If any errors were encountered, clean up and return.
     */
compFail:
    if (archComp) {
        Itk_DelArchComponent(archComp);
    }
    if (entry) {
        Tcl_DeleteHashEntry(entry);
    }
    if (path) {
        ckfree(path);
    }
    if (mergeInfo && mergeInfo->optionTable) {
        Itk_DelGenericOptTable(mergeInfo->optionTable);
        mergeInfo->optionTable = NULL;
        mergeInfo->archInfo    = NULL;
        mergeInfo->archComp    = NULL;
    }

    /*
     *  Add error info and return.
     */
    Tcl_DStringTrunc(&buffer, 0);
    Tcl_DStringAppend(&buffer, "\n    (while creating component \"", -1);
    Tcl_DStringAppend(&buffer, argv[1], -1);
    Tcl_DStringAppend(&buffer, "\"", -1);
    if (obj->accessCmd) {
        Tcl_DStringAppend(&buffer, " for widget \"", -1);
        Tcl_DStringAppend(&buffer,
            Tcl_GetCommandPath(interp, obj->accessCmd),-1);
        Tcl_DStringAppend(&buffer, "\"", -1);
    }
    Tcl_DStringAppend(&buffer, ")", -1);
    Tcl_AddErrorInfo(interp, Tcl_DStringValue(&buffer));

    Tcl_DStringFree(&buffer);
    Tcl_DStringFree(&objName);
    return TCL_ERROR;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchCompDeleteCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::itk_component
 *  method.  Removes an existing component widget from a mega-widget,
 *  and removes any configuration options associated with it.
 *
 *      itk_component delete <name> ?<name> <name>...?
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchCompDeleteCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    Itcl_Namespace classNs;
    Itcl_Class *cdefn;
    Itcl_Object *obj;
    ArchInfo *info;
    int i;
    Tcl_HashEntry *entry;
    Tcl_HashSearch place;
    Itcl_ListElem *elem;
    ArchComponent *archComp;
    ArchOption *archOpt;
    ArchOptionPart *optPart;

    /*
     *  Get the Archetype info associated with this widget.
     */
    if (Itcl_GetClassContext(interp,&classNs,&cdefn,&obj) != TCL_OK || !obj) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "cannot access components without ",
            "an object context",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (Itk_GetArchInfo(interp,obj,&info) != TCL_OK) {
        return TCL_ERROR;
    }

    /*
     *  Scan through the list of component names and delete each
     *  one.  Make sure that each component exists.
     */
    for (i=1; i < argc; i++) {
        entry = Tcl_FindHashEntry(&info->components, argv[i]);
        if (!entry) {
            Tcl_AppendResult(interp, "name \"", argv[i],
                "\" is not a component",
                (char*)NULL);
            return TCL_ERROR;
        }
        archComp = (ArchComponent*)Tcl_GetHashValue(entry);
        Tcl_DeleteHashEntry(entry);

        entry = Tcl_FirstHashEntry(&info->options, &place);
        while (entry) {
            archOpt = (ArchOption*)Tcl_GetHashValue(entry);
            elem = Itcl_FirstListElem(&archOpt->parts);
            while (elem) {
                optPart = (ArchOptionPart*)Itcl_GetListValue(elem);
                if (optPart->from == (ClientData)archComp) {
                    Itk_DelOptionPart(optPart);
                    elem = Itcl_DeleteListElem(elem);
                }
                else {
                    elem = Itcl_NextListElem(elem);
                }
            }
            entry = Tcl_NextHashEntry(&place);
        }

        if (archComp->tkwin && info->tkwin) {
            Itk_SetMegaWidget(archComp->tkwin, (Tk_Window)NULL);
        }

        Itk_DelArchComponent(archComp);
    }
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchOptKeepCmd()
 *
 *  Invoked by [incr Tcl] to handle the "keep" command in the itk
 *  option parser.  Integrates a list of component configuration options
 *  into a mega-widget, so that whenever the mega-widget is updated,
 *  the component will be updated as well.
 *
 *  Handles the following syntax:
 *
 *      keep <option> ?<option>...?
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchOptKeepCmd(clientData, interp, argc, argv)
    ClientData clientData;   /* option merging info record */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    ArchMergeInfo *mergeInfo = (ArchMergeInfo*)clientData;
    int status = TCL_OK;

    int i;
    Tcl_HashEntry *entry;
    GenericConfigOpt *opt;
    ArchOption *archOpt;
    ArchOptionPart *optPart;
    Tcl_DString code;
    char *cmdStr;

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
            argv[0], " option ?option...?\"",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Make sure that this command is being accessed in the
     *  proper context.  The merge info record should be set up
     *  properly.
     */
    if (!mergeInfo->archInfo || !mergeInfo->optionTable) {
        Tcl_AppendResult(interp, "improper usage: \"", argv[0],
            "\" should only be accessed via itk_component",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Scan through all of the options on the list, and make
     *  sure that they are valid options for this component.
     *  Integrate them into the option info for the mega-widget.
     */
    Tcl_DStringInit(&code);

    for (i=1; i < argc; i++) {
        entry = Tcl_FindHashEntry(mergeInfo->optionTable, argv[i]);
        if (!entry) {
            Tcl_AppendResult(interp, "option not recognized: ", argv[i],
                (char*)NULL);
            status = TCL_ERROR;
            break;
        }
        opt = (GenericConfigOpt*)Tcl_GetHashValue(entry);

        /*
         *  If this option has already been integrated, then
         *  remove it and start again.
         */
        Itk_IgnoreArchOptionPart(mergeInfo->archInfo, opt);

        /*
         *  Build a command prefix that can be used to apply changes
         *  to this option for this component.
         */
        Tcl_DStringTrunc(&code, 0);
        Tcl_DStringAppendElement(&code,
            Tcl_GetCommandPath(interp, mergeInfo->archComp->accessCmd));
        Tcl_DStringAppendElement(&code, "configure");
        Tcl_DStringAppendElement(&code, argv[i]);

        cmdStr = ckalloc((unsigned)(Tcl_DStringLength(&code)+1));
        strcpy(cmdStr, Tcl_DStringValue(&code));

        optPart = Itk_CreateOptionPart(interp, (ClientData)cmdStr,
            Itk_PropagateOption, (Itcl_DeleteProc*)free,
            (ClientData)mergeInfo->archComp);

        if (Itk_AddOptionPart(interp, mergeInfo->archInfo, opt->switchName,
            opt->resName, opt->resClass, opt->init, opt->value,
            optPart, &archOpt) == TCL_OK) {
            opt->integrated = archOpt;
            opt->optPart    = optPart;
        } else {
            Itk_DelOptionPart(optPart);
            status = TCL_ERROR;
            break;
        }
    }

    Tcl_DStringFree(&code);
    return status;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchOptIgnoreCmd()
 *
 *  Invoked by [incr Tcl] to handle the "ignore" command in the itk
 *  option parser.  Removes a list of component configuration options
 *  from a mega-widget.  This negates the action of previous "keep"
 *  and "rename" commands.
 *
 *  Handles the following syntax:
 *
 *      ignore <option> ?<option>...?
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchOptIgnoreCmd(clientData, interp, argc, argv)
    ClientData clientData;   /* option merging info record */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    ArchMergeInfo *mergeInfo = (ArchMergeInfo*)clientData;

    int i;
    Tcl_HashEntry *entry;
    GenericConfigOpt *opt;

    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
            argv[0], " option ?option...?\"",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Make sure that this command is being accessed in the
     *  proper context.  The merge info record should be set up
     *  properly.
     */
    if (!mergeInfo->archInfo || !mergeInfo->optionTable) {
        Tcl_AppendResult(interp, "improper usage: \"", argv[0],
            "\" should only be accessed via itk_component",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Scan through all of the options on the list, and make
     *  sure that they are valid options for this component.
     *  Remove them from the mega-widget.
     */
    for (i=1; i < argc; i++) {
        entry = Tcl_FindHashEntry(mergeInfo->optionTable, argv[i]);
        if (!entry) {
            Tcl_AppendResult(interp, "option not recognized: ", argv[i],
                (char*)NULL);
            return TCL_ERROR;
        }
        opt = (GenericConfigOpt*)Tcl_GetHashValue(entry);

        /*
         *  If this option has already been integrated, then
         *  remove it.  Otherwise, ignore it.
         */
        Itk_IgnoreArchOptionPart(mergeInfo->archInfo, opt);
    }
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchOptRenameCmd()
 *
 *  Invoked by [incr Tcl] to handle the "rename" command in the itk
 *  option parser.  Integrates one configuration option into a
 *  mega-widget, using a different name for the option.  Whenever the
 *  mega-widget option is updated, the renamed option will be updated
 *  as well.  Handles the following syntax:
 *
 *      rename <oldSwitch> <newSwitch> <resName> <resClass>
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchOptRenameCmd(clientData, interp, argc, argv)
    ClientData clientData;   /* option merging info record */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    ArchMergeInfo *mergeInfo = (ArchMergeInfo*)clientData;
    int status = TCL_OK;

    Tcl_HashEntry *entry;
    GenericConfigOpt *opt;
    ArchOption *archOpt;
    ArchOptionPart *optPart;
    Tcl_DString code;
    char *cmdStr;

    if (argc != 5) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
            argv[0], " oldSwitch newSwitch resourceName resourceClass\"",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Make sure that this command is being accessed in the
     *  proper context.  The merge info record should be set up
     *  properly.
     */
    if (!mergeInfo->archInfo || !mergeInfo->optionTable) {
        Tcl_AppendResult(interp, "improper usage: \"", argv[0],
            "\" should only be accessed via itk_component",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Make sure that the resource name and resource class look good.
     */
    if (!islower(*argv[3])) {
        Tcl_AppendResult(interp, "bad resource name \"", argv[3],
            "\": should start with a lower case letter",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (!isupper(*argv[4])) {
        Tcl_AppendResult(interp, "bad resource class \"", argv[4],
            "\": should start with an upper case letter",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Make sure that the specified switch exists in the widget.
     */
    entry = Tcl_FindHashEntry(mergeInfo->optionTable, argv[1]);
    if (!entry) {
        Tcl_AppendResult(interp, "option not recognized: ", argv[1],
            (char*)NULL);
        return TCL_ERROR;
    }
    opt = (GenericConfigOpt*)Tcl_GetHashValue(entry);

    /*
     *  If this option has already been integrated, then
     *  remove it and start again.
     */
    Itk_IgnoreArchOptionPart(mergeInfo->archInfo, opt);

    /*
     *  Build a command prefix that can be used to apply changes
     *  to this option for this component.
     */
    Tcl_DStringInit(&code);
    Tcl_DStringAppendElement(&code,
        Tcl_GetCommandPath(interp, mergeInfo->archComp->accessCmd));
    Tcl_DStringAppendElement(&code, "configure");
    Tcl_DStringAppendElement(&code, argv[1]);

    cmdStr = ckalloc((unsigned)(Tcl_DStringLength(&code)+1));
    strcpy(cmdStr, Tcl_DStringValue(&code));

    optPart = Itk_CreateOptionPart(interp, (ClientData)cmdStr,
        Itk_PropagateOption, (Itcl_DeleteProc*)free,
        (ClientData)mergeInfo->archComp);

    /*
     *  Merge this option into the mega-widget with a new name.
     */
    if (Itk_AddOptionPart(interp, mergeInfo->archInfo, argv[2],
        argv[3], argv[4], opt->init, opt->value, optPart,
        &archOpt) == TCL_OK) {
        opt->integrated = archOpt;
        opt->optPart    = optPart;
    } else {
        Itk_DelOptionPart(optPart);
        status = TCL_ERROR;
    }

    Tcl_DStringFree(&code);
    return status;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchOptUsualCmd()
 *
 *  Invoked by [incr Tcl] to handle the "usual" command in the itk
 *  option parser.  Looks for a set of "usual" option-handling commands
 *  associated with the given tag or component class and then evaluates
 *  the commands in the option parser namespace.  This keeps the user
 *  from having to type a bunch of "keep" and "rename" commands for
 *  each component widget.
 *
 *  Handles the following syntax:
 *
 *      usual ?<tag>?
 *
 *  If the <tag> is not specified, then the class name for the
 *  component is used as the tag name.
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchOptUsualCmd(clientData, interp, argc, argv)
    ClientData clientData;   /* option merging info record */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    ArchMergeInfo *mergeInfo = (ArchMergeInfo*)clientData;

    char *tag, *code;
    Tcl_HashEntry *entry;

    if (argc > 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
            argv[0], " ?tag?\"",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Make sure that this command is being accessed in the
     *  proper context.  The merge info record should be set up
     *  properly.
     */
    if (!mergeInfo->archInfo || !mergeInfo->optionTable) {
        Tcl_AppendResult(interp, "improper usage: \"", argv[0],
            "\" should only be accessed via itk_component",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  If a tag name was specified, then use this to look up
     *  the "usual" code.  Otherwise, use the class name for
     *  the component widget.
     */
    if (argc == 2) {
        tag = argv[1];
    } else {
        tag = Tk_Class(mergeInfo->archComp->tkwin);
    }

    /*
     *  Look for some code associated with the tag and evaluate
     *  it in the current context.
     */
    entry = Tcl_FindHashEntry(&mergeInfo->usualCode, tag);
    if (entry) {
        code = (char*)Tcl_GetHashValue(entry);
        return Tcl_Eval(interp, code);
    }

    Tcl_AppendResult(interp, "can't find usual code for tag \"",
        tag, "\"", (char*)NULL);
    return TCL_ERROR;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_UsualCmd()
 *
 *  Invoked by [incr Tcl] to handle the "usual" command in the ::itk
 *  namespace.  Used to query or set the option-handling code associated
 *  with a widget class or arbitrary tag name.  This code is later
 *  used by the "usual" command in the "itk::option-parser" namespace.
 *
 *  Handles the following syntax:
 *
 *      usual <tag> ?<code>?
 *
 *  If the <code> is not specified, then this returns the current
 *  code associated with <tag>, or an empty string if <tag> is not
 *  recognized.  Otherwise, it sets the code fragment for <tag> to
 *  <code>.
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
int
Itk_UsualCmd(clientData, interp, argc, argv)
    ClientData clientData;   /* option merging info record */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    ArchMergeInfo *mergeInfo = (ArchMergeInfo*)clientData;

    char *code;
    Tcl_HashEntry *entry;
    int newEntry;

    if (argc < 2 || argc > 3) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
            argv[0], " tag ?commands?\"",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  If a code fragment was specified, then save it in the
     *  hash table for "usual" code.
     */
    if (argc == 3) {
        entry = Tcl_CreateHashEntry(&mergeInfo->usualCode, argv[1], &newEntry);
        if (!newEntry) {
            code = (char*)Tcl_GetHashValue(entry);
            ckfree(code);
        }
        code = (char*)ckalloc((unsigned)(strlen(argv[2])+1));
        strcpy(code, argv[2]);
        Tcl_SetHashValue(entry, (ClientData)code);

        return TCL_OK;
    }

    /*
     *  Otherwise, look for a code fragment with the specified tag.
     */
    entry = Tcl_FindHashEntry(&mergeInfo->usualCode, argv[1]);
    if (entry) {
        code = (char*)Tcl_GetHashValue(entry);
        Tcl_SetResult(interp, code, TCL_STATIC);
    }
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchInitCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::itk_initialize
 *  method.  This method should be called out in the constructor for
 *  each mega-widget class, to build the composite option list at
 *  each class level.  Handles the following syntax:
 *
 *      itk_initialize ?-option val -option val...?
 *
 *  Integrates any class-based options into the composite option list,
 *  handles option settings from the command line, and then configures
 *  all options to have the proper initial value.
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchInitCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    Itcl_Namespace classNs;
    Itcl_Class *cdefn;
    Itcl_Object *obj;
    ArchInfo *info;

    int i, status;
    ItkClassOption *opt;
    ItkClassOptTable *optTable;
    Itcl_ListElem *part;
    ArchOption *archOpt;
    ArchOptionPart *optPart;
    Itcl_Namespace context;
    char *val;

    if (Itcl_GetClassContext(interp,&classNs,&cdefn,&obj) != TCL_OK || !obj) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "improper usage: should be \"object ",
            argv[0], " ?-option value -option value...?\"",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (Itk_GetArchInfo(interp,obj,&info) != TCL_OK) {
        return TCL_ERROR;
    }

    /*
     *  See what class is being initialized by getting the namespace
     *  for the calling context.
     */
    context = Itcl_GetCallingNamesp(interp,1);
    if (Itcl_IsClass(context)) {
        cdefn = (Itcl_Class*)Itcl_GetNamespData(context);
    }

    /*
     *  Integrate all class-based options for the current class
     *  context into the composite option list.
     */
    optTable = Itk_FindClassOptTable(cdefn);
    if (optTable) {
        for (i=0; i < optTable->order.len; i++) {
            opt = (ItkClassOption*)Tcl_GetHashValue(optTable->order.list[i]);

            optPart = Itk_FindArchOptionPart(info, opt->switchName,
                (ClientData)cdefn);

            if (!optPart) {
                optPart = Itk_CreateOptionPart(interp, (ClientData)opt,
                    Itk_ConfigClassOption, (Itcl_DeleteProc*)NULL,
                    (ClientData)cdefn);

                if (Itk_AddOptionPart(interp, info, opt->switchName,
                    opt->resName, opt->resClass, opt->init, (char*)NULL,
                    optPart, &archOpt) != TCL_OK) {
                    Itk_DelOptionPart(optPart);
                    return TCL_ERROR;
                }
            }
        }
    }

    /*
     *  If any option values were specified on the command line,
     *  override the current option settings.  Treat these settings
     *  in the proper invocation context.  The class constructor
     *  should be one level up, so the real invocation context
     *  (where the object is being created) is two levels up.
     */
    if (argc > 1) {
        for (argc--,argv++; argc > 0; argc-=2, argv+=2) {
            if (argc < 2) {
                Tcl_AppendResult(interp, "value for \"", *argv,
                    "\" missing",
                    (char*)NULL);
                return TCL_ERROR;
            }
            if (Itk_ArchConfigOption(interp, info,
                argv[0], argv[1], 2) != TCL_OK) {
                return TCL_ERROR;
            }
        }
    }

    /*
     *  If this is most-specific class, then finish constructing
     *  the mega-widget:
     *
     *  Scan through all options in the composite list and
     *  look for any that have been set but not initialized.
     *  Invoke the parts of uninitialized options to propagate
     *  changes and update the widget.
     */
    if (obj->cdefn == cdefn) {
        for (i=0; i < info->order.len; i++) {
            archOpt = (ArchOption*)Tcl_GetHashValue(info->order.list[i]);

            if ((archOpt->flags & ITK_ARCHOPT_INIT) == 0) {
                val = Tcl_GetVar2(interp, "itk_option", archOpt->switchName, 0);

                if (!val) {
                    Itk_ArchOptAccessError(interp, info, archOpt);
                    return TCL_ERROR;
                }

                part = Itcl_FirstListElem(&archOpt->parts);
                while (part) {
                    optPart = (ArchOptionPart*)Itcl_GetListValue(part);
                    status  = (*optPart->configProc)(interp, obj,
                        optPart->clientData, val);

                    if (status != TCL_OK) {
                        Itk_ArchOptConfigError(interp, info, archOpt);
                    }
                    part = Itcl_NextListElem(part);
                }
                archOpt->flags |= ITK_ARCHOPT_INIT;
            }
        }
    }

    Tcl_ResetResult(interp);
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchOptionCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::itk_option
 *  method.  Handles the following options:
 *
 *      itk_option define <switch> <resName> <resClass> <init> ?<config>?
 *      itk_option add <name> ?<name>...?
 *      itk_option remove <name> ?<name>...?
 *
 *  These commands customize the options list of a specific widget.
 *  They are similar to the "itk_option" ensemble in the class definition
 *  parser, but manipulate a single instance instead of an entire class.
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchOptionCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    char c;
    int length;

    /*
     *  Check arguments and handle the various options...
     */
    if (argc < 2) {
        Tcl_AppendResult(interp, "wrong # args: should be one of...\n",
            "  ", argv[0], " add name ?name name...?\n",
            "  ", argv[0], " define -switch resourceName resourceClass init ?config?\n",
            "  ", argv[0], " remove name ?name name...?",
            (char*)NULL);
        return TCL_ERROR;
    }
    c = *argv[1];
    length = strlen(argv[1]);

    /*
     *  Handle:  itk_option add...
     */
    if (c == 'a' && strncmp(argv[1], "add", length) == 0) {
        if (argc < 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                argv[0], " add name ?name name...?\"",
                (char*)NULL);
            return TCL_ERROR;
        }
        return Itk_ArchOptionAddCmd(dummy, interp, argc-1, argv+1);
    }

    /*
     *  Handle:  itk_option remove...
     */
    else if (c == 'r' && strncmp(argv[1], "remove", length) == 0) {
        if (argc < 3) {
            Tcl_AppendResult(interp, "wrong # args: should be \"",
                argv[0], " remove name ?name name...?\"",
                (char*)NULL);
            return TCL_ERROR;
        }
        return Itk_ArchOptionRemoveCmd(dummy, interp, argc-1, argv+1);
    }

    /*
     *  Handle:  itk_option define...
     */
    else if (c == 'd' && strncmp(argv[1], "define", length) == 0) {
        Tcl_AppendResult(interp, "can only ", argv[1],
            " options at the class level\n",
            "(move this command into the class definition)",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Flag any errors.
     */
    Tcl_AppendResult(interp, "bad option \"", argv[1],
        "\": should be one of...\n",
        "  ", argv[0], " add name ?name name...?\n",
        "  ", argv[0], " define -switch resourceName resourceClass init ?config?\n",
        "  ", argv[0], " remove name ?name name...?",
        (char*)NULL);
    return TCL_ERROR;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchOptionAddCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::itk_option add
 *  method.  Finds an option within a class definition or belonging to
 *  a component widget and adds it into the option list for this widget.
 *  If the option is already on the list, this method does nothing.
 *  Handles the following syntax:
 *
 *      itk_option add <name> ?<name> <name>...?
 *
 *      where <name> is one of:
 *        class::option
 *        component.option
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchOptionAddCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    Itcl_Namespace classNs, ns;
    Itcl_Class *cdefn, *cd;
    Itcl_Object *obj;
    ArchInfo *info;

    int i;
    char *head, *tail, *sep, tmp, *cmdStr;
    ItkClassOption *opt;
    GenericConfigOpt *generic;
    ArchOption *archOpt;
    ArchOptionPart *optPart;
    ArchComponent *archComp;
    Tcl_HashEntry *entry;
    Tcl_DString buffer;

    /*
     *  Get the Archetype info associated with this widget.
     */
    if (Itcl_GetClassContext(interp,&classNs,&cdefn,&obj) != TCL_OK || !obj) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "cannot access options without ",
            "an object context",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (Itk_GetArchInfo(interp,obj,&info) != TCL_OK) {
        return TCL_ERROR;
    }

    /*
     *  Scan through the list of options and locate each one.
     *  If it is not already on the option part list, add it.
     */
    for (i=1; i < argc; i++) {
        Itcl_ParseNamespPath(argv[i], &head, &tail);

        /*
         *  HANDLE:  class::option
         */
        if (head) {
            if (Itcl_FindClass(interp, head, &ns) != TCL_OK) {
                return TCL_ERROR;
            }
            if (ns == NULL) {
                Tcl_AppendResult(interp, "class \"", head,
                    "\" not found in context \"",
                    Itcl_GetNamespPath(classNs), "\"",
                    (char*)NULL);
                return TCL_ERROR;
            }
            cd = (Itcl_Class*)Itcl_GetNamespData(ns);

            opt = Itk_FindClassOption(cd, tail);
            if (!opt) {
                Tcl_AppendResult(interp, "option \"", tail,
                    "\" not defined in class \"",
                    Itcl_GetNamespPath(classNs), "\"",
                    (char*)NULL);
                return TCL_ERROR;
            }

            optPart = Itk_FindArchOptionPart(info, opt->switchName,
                (ClientData)cd);

            if (!optPart) {
                optPart = Itk_CreateOptionPart(interp, (ClientData)opt,
                    Itk_ConfigClassOption, (Itcl_DeleteProc*)NULL,
                    (ClientData)cd);

                if (Itk_AddOptionPart(interp, info, opt->switchName,
                    opt->resName, opt->resClass, opt->init, (char*)NULL,
                    optPart, &archOpt) != TCL_OK) {
                    Itk_DelOptionPart(optPart);
                    return TCL_ERROR;
                }
            }
        }

        /*
         *  HANDLE:  component.option
         */
        else if ((sep=strstr(argv[i],".")) != NULL) {
            tmp = *sep;
            *sep = '\0';
            head = argv[i];
            tail = sep+1;

            entry = Tcl_FindHashEntry(&info->components, head);
            if (!entry) {
                Tcl_AppendResult(interp, "name \"", head,
                    "\" is not a component",
                    (char*)NULL);
                *sep = tmp;
                return TCL_ERROR;
            }
            *sep = tmp;
            archComp = (ArchComponent*)Tcl_GetHashValue(entry);

            generic = Itk_CreateGenericOpt(interp, tail, archComp->accessCmd);
            if (!generic) {
                char msg[256];
                sprintf(msg, "\n    (while adding option \"%.100s\")", argv[i]);
                Tcl_AddErrorInfo(interp, msg);
                return TCL_ERROR;
            }

            optPart = Itk_FindArchOptionPart(info, generic->switchName,
                (ClientData)archComp);

            if (!optPart) {
                Tcl_DStringInit(&buffer);
                Tcl_DStringAppendElement(&buffer,
                    Tcl_GetCommandPath(interp, archComp->accessCmd));
                Tcl_DStringAppendElement(&buffer, "configure");
                Tcl_DStringAppendElement(&buffer, generic->switchName);

                cmdStr = ckalloc((unsigned)(Tcl_DStringLength(&buffer)+1));
                strcpy(cmdStr, Tcl_DStringValue(&buffer));
                Tcl_DStringFree(&buffer);

                optPart = Itk_CreateOptionPart(interp, (ClientData)cmdStr,
                    Itk_PropagateOption, (Itcl_DeleteProc*)free,
                    (ClientData)archComp);

                if (Itk_AddOptionPart(interp, info, generic->switchName,
                    generic->resName, generic->resClass, generic->init,
                    generic->value, optPart, &archOpt) != TCL_OK) {
                    Itk_DelOptionPart(optPart);
                    Itk_DelGenericOpt(generic);
                    return TCL_ERROR;
                }
            }
            Itk_DelGenericOpt(generic);
        }

        /*
         *  Anything else is an error.
         */
        else {
            Tcl_AppendResult(interp, "bad option \"", argv[i],
                "\": should be one of...\n",
                "  class::option\n",
                "  component.option",
                (char*)NULL);
            return TCL_ERROR;
        }
    }
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchOptionRemoveCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::itk_option remove
 *  method.  Finds an option within a class definition or belonging to
 *  a component widget and removes it from the option list for this widget.
 *  If the option has already been removed from the list, this method does
 *  nothing.  Handles the following syntax:
 *
 *      itk_option remove <name> ?<name> <name>...?
 *
 *      where <name> is one of:
 *        class::option
 *        component.option
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchOptionRemoveCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    Itcl_Namespace classNs, ns;
    Itcl_Class *cdefn, *cd;
    Itcl_Object *obj;
    ArchInfo *info;

    int i;
    char *head, *tail, *sep, tmp;
    ItkClassOption *opt;
    GenericConfigOpt *generic;
    ArchComponent *archComp;
    Tcl_HashEntry *entry;

    /*
     *  Get the Archetype info associated with this widget.
     */
    if (Itcl_GetClassContext(interp,&classNs,&cdefn,&obj) != TCL_OK || !obj) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "cannot access options without ",
            "an object context",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (Itk_GetArchInfo(interp,obj,&info) != TCL_OK) {
        return TCL_ERROR;
    }

    /*
     *  Scan through the list of options and locate each one.
     *  If it is on the option list, remove it.
     */
    for (i=1; i < argc; i++) {
        Itcl_ParseNamespPath(argv[i], &head, &tail);

        /*
         *  HANDLE:  class::option
         */
        if (head) {
            if (Itcl_FindClass(interp, head, &ns) != TCL_OK) {
                return TCL_ERROR;
            }
            if (ns == NULL) {
                Tcl_AppendResult(interp, "class \"", head,
                    "\" not found in context \"",
                    Itcl_GetNamespPath(classNs), "\"",
                    (char*)NULL);
                return TCL_ERROR;
            }
            cd = (Itcl_Class*)Itcl_GetNamespData(ns);

            opt = Itk_FindClassOption(cd, tail);
            if (!opt) {
                Tcl_AppendResult(interp, "option \"", tail,
                    "\" not defined in class \"",
                    Itcl_GetNamespPath(classNs), "\"",
                    (char*)NULL);
                return TCL_ERROR;
            }

            Itk_RemoveArchOptionPart(info, opt->switchName,
                (ClientData)cd);
        }

        /*
         *  HANDLE:  component.option
         */
        else if ((sep=strstr(argv[i],".")) != NULL) {
            tmp = *sep;
            *sep = '\0';
            head = argv[i];
            tail = sep+1;

            entry = Tcl_FindHashEntry(&info->components, head);
            if (!entry) {
                Tcl_AppendResult(interp, "name \"", head,
                    "\" is not a component",
                    (char*)NULL);
                *sep = tmp;
                return TCL_ERROR;
            }
            *sep = tmp;
            archComp = (ArchComponent*)Tcl_GetHashValue(entry);

            generic = Itk_CreateGenericOpt(interp, tail, archComp->accessCmd);
            if (!generic) {
                char msg[256];
                sprintf(msg, "\n    (while removing option \"%.100s\")",
                    argv[i]);
                Tcl_AddErrorInfo(interp, msg);
                return TCL_ERROR;
            }

            Itk_RemoveArchOptionPart(info, generic->switchName,
                (ClientData)archComp);

            Itk_DelGenericOpt(generic);
        }

        /*
         *  Anything else is an error.
         */
        else {
            Tcl_AppendResult(interp, "bad option \"", argv[i],
                "\": should be one of...\n",
                "  class::option\n",
                "  component.option",
                (char*)NULL);
            return TCL_ERROR;
        }
    }
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchCompAccessCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::component method.
 *  Finds the requested component and invokes the <command> as a method
 *  on that component.
 *
 *  Handles the following syntax:
 *
 *      component
 *      component <name>
 *      component <name> <command> ?<arg> <arg>...?
 *
 *  With no arguments, this command returns the names of components
 *  that can be accessed from the current context.  Note that components
 *  respect public/protected/private declarations, so private and
 *  protected components may not be accessible from all namespaces.
 *
 *  If a component name is specified, then this command returns the
 *  window name for that component.
 *
 *  If a series of arguments follow the component name, they are treated
 *  as a method invocation, and dispatched to the component.
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchCompAccessCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    Itcl_Namespace classNs;
    Itcl_Class *cdefn;
    Itcl_Object *obj;
    Tcl_HashEntry *entry;
    Tcl_HashSearch place;
    ArchInfo *info;
    ArchComponent *archComp;
    Itcl_ActiveNamespace nsToken;
    Namespace *ownerPtr, *accessPtr;
    char *name, *argv1, *cmdStr, *val;
    int status;

    if (Itcl_GetClassContext(interp,&classNs,&cdefn,&obj) != TCL_OK || !obj) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "improper usage: should be \"object ",
            argv[0], " ?name option arg arg...?\"",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (Itk_GetArchInfo(interp,obj,&info) != TCL_OK) {
        return TCL_ERROR;
    }

    accessPtr = (Namespace*)Itcl_GetCallingNamesp(interp,1);

    /*
     *  With no arguments, return a list of components that can be
     *  accessed from the calling scope.
     */
    if (argc == 1) {
        entry = Tcl_FirstHashEntry(&info->components, &place);
        while (entry) {
            archComp = (ArchComponent*)Tcl_GetHashValue(entry);
            ownerPtr = (Namespace*)Itcl_GetCommandNamesp(archComp->accessCmd);

            if (ItclCanAccessNamesp(ownerPtr, accessPtr,
                archComp->protection)) {
                name = Tcl_GetHashKey(&info->components, entry);
                Tcl_AppendElement(interp, name);
            }
            entry = Tcl_NextHashEntry(&place);
        }
        return TCL_OK;
    }

    /*
     *  Make sure the requested component exists.
     */
    entry = Tcl_FindHashEntry(&info->components, argv[1]);
    if (entry) {
        archComp = (ArchComponent*)Tcl_GetHashValue(entry);
    } else {
        archComp = NULL;
    }

    if (archComp == NULL) {
        Tcl_AppendResult(interp, "name \"", argv[1], "\" is not a component",
            (char*)NULL);
        return TCL_ERROR;
    }

    ownerPtr = (Namespace*)Itcl_GetCommandNamesp(archComp->accessCmd);
    if (!ItclCanAccessNamesp(ownerPtr, accessPtr, archComp->protection)) {
        Tcl_AppendResult(interp, "can't access component \"", argv[1],
            "\" from namespace \"",
            Itcl_GetNamespPath((Itcl_Namespace)accessPtr), "\"",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  If only the component name is specified, then return the
     *  window name for this component.
     */
    if (argc == 2) {
        val = Tcl_GetVar2(interp, "itk_component", argv[1], 0);
        if (!val) {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "internal error: cannot access ",
                "itk_component(", argv[1], ")",
                (char*)NULL);
            if (obj->accessCmd) {
                Tcl_AppendResult(interp, " in widget \"",
                    Tcl_GetCommandPath(interp, obj->accessCmd), "\"",
                    (char*)NULL);
            }
            return TCL_ERROR;
        }
        Tcl_SetResult(interp, val, TCL_VOLATILE);
        return TCL_OK;
    }

    /*
     *  Otherwise, treat the rest of the command line as a method
     *  invocation on the requested component.  Invoke the remaining
     *  command-line arguments as a method for that component.
     */
    nsToken = Itcl_ActivateNamesp(interp,
        Itcl_GetCommandNamesp(archComp->accessCmd));

    argv1 = argv[1];
    argv[1] = Tcl_GetCommandName(interp, archComp->accessCmd);
    cmdStr = Tcl_Merge(argc-1,argv+1);

    status = Itcl_EvalArgs(interp, cmdStr, (char*)NULL, argc-1, argv+1);

    ckfree(cmdStr);
    argv[1] = argv1;

    Itcl_DeactivateNamesp(interp, nsToken);

    return status;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchConfigureCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::configure method.
 *  Mimics the usual Tk "configure" method for Archetype mega-widgets.
 *
 *      configure
 *      configure -name
 *      configure -name value ?-name value ...?
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchConfigureCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    int i;
    Itcl_Namespace classNs;
    Itcl_Class *cdefn;
    Itcl_Object *obj;
    ArchInfo *info;
    Tcl_HashEntry *entry;
    ArchOption *archOpt;
    char *val;
    Tcl_DString buffer;

    if (Itcl_GetClassContext(interp,&classNs,&cdefn,&obj) != TCL_OK || !obj) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "improper usage: should be \"object ",
            argv[0], " ?-option? ?value -option value...?\"",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (Itk_GetArchInfo(interp,obj,&info) != TCL_OK) {
        return TCL_ERROR;
    }

    /*
     *  If there are no extra arguments, then return a list of all
     *  known configuration options.  Each option has the form:
     *    {name resName resClass init value}
     */
    if (argc == 1) {
        Tcl_DStringInit(&buffer);

        for (i=0; i < info->order.len; i++) {
            archOpt = (ArchOption*)Tcl_GetHashValue(info->order.list[i]);
            val = Tcl_GetVar2(interp, "itk_option", archOpt->switchName, 0);
            if (!val) {
                Itk_ArchOptAccessError(interp, info, archOpt);
                Tcl_DStringFree(&buffer);
                return TCL_ERROR;
            }

            Tcl_DStringStartSublist(&buffer);
            Tcl_DStringAppendElement(&buffer, archOpt->switchName);
            Tcl_DStringAppendElement(&buffer, archOpt->resName);
            Tcl_DStringAppendElement(&buffer, archOpt->resClass);
            Tcl_DStringAppendElement(&buffer, archOpt->init);
            Tcl_DStringAppendElement(&buffer, val);
            Tcl_DStringEndSublist(&buffer);
        }
        Tcl_DStringResult(interp, &buffer);
        Tcl_DStringFree(&buffer);
        return TCL_OK;
    }

    /*
     *  If there is just one argument, then query the information
     *  for that one argument and return:
     *    {name resName resClass init value}
     */
    else if (argc == 2) {
        entry = Tcl_FindHashEntry(&info->options, argv[1]);
        if (!entry) {
            Tcl_AppendResult(interp, "unknown option \"", argv[1], "\"",
                (char*)NULL);
            return TCL_ERROR;
        }

        archOpt = (ArchOption*)Tcl_GetHashValue(entry);
        val = Tcl_GetVar2(interp, "itk_option", archOpt->switchName, 0);
        if (!val) {
            Itk_ArchOptAccessError(interp, info, archOpt);
            return TCL_ERROR;
        }

        Tcl_AppendElement(interp, archOpt->switchName);
        Tcl_AppendElement(interp, archOpt->resName);
        Tcl_AppendElement(interp, archOpt->resClass);
        Tcl_AppendElement(interp, archOpt->init);
        Tcl_AppendElement(interp, val);

        return TCL_OK;
    }

    /*
     *  Otherwise, it must be a series of "-option value" assignments.
     *  Look up each option and assign the new value.  Perform the
     *  "configure" operation in the context of the calling scope,
     *  so options that are sensitive to scope will get the proper
     *  value.
     */
    for (argc--,argv++; argc > 0; argc-=2, argv+=2) {
        if (argc < 2) {
            Tcl_AppendResult(interp, "value for \"", *argv, "\" missing",
                (char*)NULL);
            return TCL_ERROR;
        }
        if (Itk_ArchConfigOption(interp, info, argv[0],argv[1], 1) != TCL_OK) {
            return TCL_ERROR;
        }
    }

    Tcl_ResetResult(interp);
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchCgetCmd()
 *
 *  Invoked by [incr Tcl] to handle the itk::Archetype::cget method.
 *  Mimics the usual Tk "cget" method for Archetype mega-widgets.
 *
 *      cget -name
 *
 *  Returns TCL_OK/TCL_ERROR to indicate success/failure.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_ArchCgetCmd(dummy, interp, argc, argv)
    ClientData dummy;        /* unused */
    Tcl_Interp *interp;      /* current interpreter */
    int argc;                /* number of arguments */
    char **argv;             /* argument strings */
{
    Itcl_Namespace classNs;
    Itcl_Class *cdefn;
    Itcl_Object *obj;
    ArchInfo *info;
    Tcl_HashEntry *entry;
    ArchOption *archOpt;
    char *val;

    if (Itcl_GetClassContext(interp,&classNs,&cdefn,&obj) != TCL_OK || !obj) {
        Tcl_ResetResult(interp);
        Tcl_AppendResult(interp, "improper usage: should be \"object ",
            argv[0], " ?-option? ?value -option value...?\"",
            (char*)NULL);
        return TCL_ERROR;
    }
    if (Itk_GetArchInfo(interp,obj,&info) != TCL_OK) {
        return TCL_ERROR;
    }

    if (argc != 2) {
        Tcl_AppendResult(interp, "wrong # args: should be \"",
            argv[0], " option\"",
            (char*)NULL);
        return TCL_ERROR;
    }

    /*
     *  Look up the specified option and get its current value.
     */
    entry = Tcl_FindHashEntry(&info->options, argv[1]);
    if (!entry) {
        Tcl_AppendResult(interp, "unknown option \"", argv[1], "\"",
            (char*)NULL);
        return TCL_ERROR;
    }

    archOpt = (ArchOption*)Tcl_GetHashValue(entry);
    val = Tcl_GetVar2(interp, "itk_option", archOpt->switchName, 0);
    if (!val) {
        Itk_ArchOptAccessError(interp, info, archOpt);
        return TCL_ERROR;
    }

    Tcl_SetResult(interp, val, TCL_VOLATILE);
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_PropagateOption()
 *
 *  Invoked whenever a widget-based configuration option has been
 *  configured with a new value.  Propagates the new value down to
 *  the widget by invoking the "configure" method on the widget.
 *  This causes the widget to bring itself up to date automatically.
 *
 *  Returns TCL_OK on success, or TCL_ERROR (along with an error
 *  message in the interpreter) if anything goes wrong.
 * ------------------------------------------------------------------------
 */
/* ARGSUSED */
static int
Itk_PropagateOption(interp,obj,cdata,newval)
    Tcl_Interp *interp;        /* interpreter managing the class */
    Itcl_Object *obj;          /* object being configured */
    ClientData cdata;          /* command prefix to use for configuration */
    char *newval;              /* new value for this option */
{
    char *cmdStr = (char*)cdata;
    int status;
    Tcl_DString buffer;

    Tcl_DStringInit(&buffer);
    Tcl_DStringAppend(&buffer, cmdStr, -1);
    Tcl_DStringAppendElement(&buffer, newval);

    status = Tcl_Eval(interp, Tcl_DStringValue(&buffer));

    Tcl_DStringFree(&buffer);
    return status;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchSetOption()
 *
 *  Sets a configuration option within an Archetype mega-widget.
 *  Changes the "itk_option" array to reflect the new value, but
 *  unlike Itk_ArchConfigOption(), this procedure does not update
 *  the widget by propagating changes or invoking any "config" code.
 *  It merely sets the widget state.  It is useful when a widget is
 *  first being constructed, to initialize option values.
 *
 *  NOTE:  This procedure assumes that there is a valid object context
 *    and a call frame supporting object data member access.  It is
 *    usually called from within the methods of the Archetype base
 *    class, so this is a good assumption.  If it is called anywhere
 *    else, the caller is responsible for installing the object context
 *    and setting up a call frame.
 *
 *  Returns TCL_OK on success, or TCL_ERROR (along with an error
 *  message in the interpreter) if anything goes wrong.
 * ------------------------------------------------------------------------
 */
static int
Itk_ArchSetOption(interp,info,name,value)
    Tcl_Interp *interp;        /* interpreter managing this widget */
    ArchInfo *info;            /* Archetype info */
    char *name;                /* name of configuration option */
    char *value;               /* new value for configuration option */
{
    Tcl_HashEntry *entry;
    ArchOption *archOpt;

    entry = Tcl_FindHashEntry(&info->options, name);
    if (!entry) {
        Tcl_AppendResult(interp, "unknown option \"", name, "\"",
            (char*)NULL);
        return TCL_ERROR;
    }
    archOpt = (ArchOption*)Tcl_GetHashValue(entry);

    if (!Tcl_SetVar2(interp, "itk_option", archOpt->switchName, value, 0)) {
        Itk_ArchOptAccessError(interp, info, archOpt);
        return TCL_ERROR;
    }
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchConfigOption()
 *
 *  Sets a configuration option within an Archetype mega-widget.
 *  Changes the "itk_option" array to reflect the new value, and then
 *  invokes any option parts to handle the new setting or propagate
 *  the value down to component parts.
 *
 *  NOTE:  This procedure assumes that there is a valid object context
 *    and a call frame supporting object data member access.  It is
 *    usually called from within the methods of the Archetype base
 *    class, so this is a good assumption.  If it is called anywhere
 *    else, the caller is responsible for installing the object context
 *    and setting up a call frame.
 *
 *  Returns TCL_OK on success, or TCL_ERROR (along with an error
 *  message in the interpreter) if anything goes wrong.
 * ------------------------------------------------------------------------
 */
static int
Itk_ArchConfigOption(interp,info,name,value,uplevel)
    Tcl_Interp *interp;        /* interpreter managing this widget */
    ArchInfo *info;            /* Archetype info */
    char *name;                /* name of configuration option */
    char *value;               /* new value for configuration option */
    int uplevel;               /* number of levels up for "configure" */
{
    Interp *iPtr = (Interp*)interp;

    int status;
    char *v, *lastval;
    Tcl_HashEntry *entry;
    ArchOption *archOpt;
    Itcl_ListElem *part;
    ArchOptionPart *optPart;
    Itcl_Namespace ns;
    ClientData cdata;
    Itcl_ActiveNamespace nsToken;
    CallFrame *framePtr, *savedVarFramePtr;

    /*
     *  Query the "itk_option" array to get the current setting.
     */
    entry = Tcl_FindHashEntry(&info->options, name);
    if (!entry) {
        Tcl_AppendResult(interp, "unknown option \"", name, "\"",
            (char*)NULL);
        return TCL_ERROR;
    }
    archOpt = (ArchOption*)Tcl_GetHashValue(entry);

    v = Tcl_GetVar2(interp, "itk_option", archOpt->switchName, 0);
    if (v) {
        lastval = (char*)ckalloc((unsigned)(strlen(v)+1));
        strcpy(lastval, v);
    } else {
        lastval = NULL;
    }

    /*
     *  Update the "itk_option" array with the new setting.
     */
    if (!Tcl_SetVar2(interp, "itk_option", archOpt->switchName, value, 0)) {
        Itk_ArchOptAccessError(interp, info, archOpt);
        status = TCL_ERROR;
        goto configDone;
    }

    /*
     *  Modify the call stack in the interpreter to point up
     *  the specified number of levels.
     */
    assert(uplevel >= 1);

    framePtr = iPtr->varFramePtr;
    while (framePtr && uplevel > 0) {
        framePtr = framePtr->callerVarPtr;
        uplevel--;
    }
    savedVarFramePtr = iPtr->varFramePtr;
    iPtr->varFramePtr = framePtr;

    if (framePtr == NULL) {
        ns    = (Itcl_Namespace)iPtr->globalNs;
        cdata = NULL;
    }
    else {
        ns    = (Itcl_Namespace)framePtr->activeNs;
        cdata = framePtr->activeData;
    }

    nsToken = Itcl_ActivateNamesp2(interp, ns, cdata);
    if (nsToken) {
        /*
         *  Scan through all option parts to handle the new setting.
         */
        status = TCL_OK;
        part   = Itcl_FirstListElem(&archOpt->parts);

        while (part) {
            optPart = (ArchOptionPart*)Itcl_GetListValue(part);
            status  = (*optPart->configProc)(interp, info->obj,
                optPart->clientData, value);

            if (status != TCL_OK) {
                Itk_ArchOptConfigError(interp, info, archOpt);
                break;
            }
            part = Itcl_NextListElem(part);
        }
    }
    else {
        status = TCL_ERROR;
    }
    Itcl_DeactivateNamesp(interp, nsToken);
    iPtr->varFramePtr = savedVarFramePtr;

    /*
     *  If the option configuration failed, then replace the
     *  old option value.
     */
    if (status == TCL_ERROR) {
        Tcl_SetVar2(interp, "itk_option", archOpt->switchName, lastval, 0);
    }

    archOpt->flags |= ITK_ARCHOPT_INIT;  /* option has been set */

configDone:
    if (lastval) {
        ckfree(lastval);
    }
    return status;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchOptConfigError()
 *
 *  Simply utility which adds error information after a option
 *  configuration fails.  Adds traceback information to the given
 *  interpreter.
 * ------------------------------------------------------------------------
 */
static void
Itk_ArchOptConfigError(interp,info,archOpt)
    Tcl_Interp *interp;            /* interpreter handling this object */
    ArchInfo *info;                /* info associated with mega-widget */
    ArchOption *archOpt;           /* configuration option that failed */
{
    Tcl_DString buffer;

    Tcl_DStringInit(&buffer);
    Tcl_DStringAppend(&buffer, "\n    (while configuring option \"",-1);
    Tcl_DStringAppend(&buffer, archOpt->switchName, -1);
    Tcl_DStringAppend(&buffer, "\"", -1);

    if (info->obj && info->obj->accessCmd) {
        Tcl_DStringAppend(&buffer, " for widget \"", -1);
        Tcl_DStringAppend(&buffer,
            Tcl_GetCommandPath(interp, info->obj->accessCmd),-1);
        Tcl_DStringAppend(&buffer, "\")", -1);
    }
    Tcl_AddErrorInfo(interp, Tcl_DStringValue(&buffer));
    Tcl_DStringFree(&buffer);
}


/*
 * ------------------------------------------------------------------------
 *  Itk_ArchOptAccessError()
 *
 *  Simply utility which adds error information after an option
 *  value access fails.  Adds traceback information to the given
 *  interpreter.
 * ------------------------------------------------------------------------
 */
static void
Itk_ArchOptAccessError(interp,info,archOpt)
    Tcl_Interp *interp;            /* interpreter handling this object */
    ArchInfo *info;                /* info associated with mega-widget */
    ArchOption *archOpt;           /* option that couldn't be accessed */
{
    Tcl_ResetResult(interp);

    Tcl_AppendResult(interp, "internal error: cannot access ",
        "itk_option(", archOpt->switchName, ")",
        (char*)NULL);

    if (info->obj->accessCmd) {
        Tcl_AppendResult(interp, " in widget \"",
            Tcl_GetCommandPath(interp, info->obj->accessCmd), "\"",
            (char*)NULL);
    }
}


/*
 * ------------------------------------------------------------------------
 *  Itk_GetArchInfo()
 *
 *  Finds the extra Archetype info associated with the given object.
 *  Returns TCL_OK and a pointer to the info if found.  Returns
 *  TCL_ERROR along with an error message in interp->result if not.
 * ------------------------------------------------------------------------
 */
static int
Itk_GetArchInfo(interp,obj,infoPtr)
    Tcl_Interp *interp;            /* interpreter handling this object */
    Itcl_Object *obj;              /* object with desired data */
    ArchInfo **infoPtr;            /* returns:  pointer to extra info */
{
    Tcl_HashEntry *entry = NULL;

    /*
     *  If there is any problem finding the info, return an error.
     */
    if (ObjsWithArchInfo != NULL) {
        entry = Tcl_FindHashEntry(ObjsWithArchInfo, (char*)obj);
    }
    if (!entry) {
        Tcl_AppendResult(interp, "internal error: no Archetype ",
            "information for widget",
            (char*)NULL);
        if (obj->accessCmd) {
            Tcl_AppendResult(interp, " \"",
                Tcl_GetCommandPath(interp, obj->accessCmd), "\"",
                (char*)NULL);
        }
        return TCL_ERROR;
    }

    /*
     *  Otherwise, return the requested info.
     */
    *infoPtr = (ArchInfo*)Tcl_GetHashValue(entry);
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_CreateArchComponent()
 *
 *  Creates the data representing a component widget within an Archetype
 *  mega-widget.  Each component has an access command that is used to
 *  communicate with it.  The protection level for this access command
 *  can be controlled by the usual public/protected/private declarations.
 *  Each component is registered by its symbolic name in the
 *  "itk_component" array.
 *
 *  Returns a pointer to the new record.  If anything goes wrong,
 *  this returns NULL, along with an error message in the interpreter.
 * ------------------------------------------------------------------------
 */
static ArchComponent*
Itk_CreateArchComponent(interp,info,name,accessCmd)
    Tcl_Interp *interp;            /* interpreter managing the object */
    ArchInfo *info;                /* info associated with mega-widget */
    char *name;                    /* symbolic name for this component */
    Tcl_Command accessCmd;         /* access command for component */
{
    ArchComponent *archComp;
    ArchOption *archOpt;
    char *wname, *init;
    Tk_Window tkwin;
    Tcl_HashEntry *entry;
    Tcl_HashSearch place;

    /*
     *  Save this component in the itk_component() array.
     */
    wname = Tcl_GetCommandName(interp, accessCmd);
    Tcl_SetVar2(interp, "itk_component", name, wname, 0);

    /*
     *  If the symbolic name for the component is "hull", then this
     *  is the toplevel or frame that embodies a mega-widget.  Update
     *  the Archtype info to include the window token.
     */
    tkwin = Tk_NameToWindow(interp, wname, Tk_MainWindow(interp));

    if (strcmp(name,"hull") == 0) {
        if (tkwin == NULL) {
            Tcl_AppendResult(interp, "cannot find hull window with ",
                "access command \"", wname, "\"",
                (char*)NULL);
            return NULL;
        }
        info->tkwin = tkwin;

        /*
         *  Register the object's access command as the access
         *  command for the "hull" window.  Otherwise, the frame
         *  or toplevel access command would be used in bind contexts.
         */
        Itk_SetWidgetCommand(tkwin, info->obj->accessCmd);

        /*
         *  We are now in a position to query configuration options
         *  relative to this window.  Scan through all existing options
         *  and update the initial values according to the X11 resource
         *  database.
         */
        entry = Tcl_FirstHashEntry(&info->options, &place);
        while (entry) {
            archOpt = (ArchOption*)Tcl_GetHashValue(entry);
            init = Tk_GetOption(tkwin, archOpt->resName, archOpt->resClass);
            if (init && (strcmp(init, archOpt->init) != 0)) {
                ckfree(archOpt->init);
                archOpt->init = (char*)ckalloc((unsigned)(strlen(init)+1));
                strcpy(archOpt->init, init);

                if (Itk_ArchSetOption(interp, info,
                    archOpt->switchName, init) != TCL_OK) {
                    return NULL;
                }
            }
            entry = Tcl_NextHashEntry(&place);
        }
    }

    /*
     *  Create the record to represent this component.
     */
    archComp = (ArchComponent*)ckalloc(sizeof(ArchComponent));
    archComp->protection = Itcl_ChangeCmdProtection(interp, accessCmd, 0);
    archComp->accessCmd  = accessCmd;
    archComp->tkwin      = tkwin;

    /*
     *  If the component access command is protected, we have made
     *  note of it above, and it will be protected by the "component"
     *  method.  The command itself, however, must be public so that
     *  it can be accessed by the mega-widget.
     */
    if (archComp->protection != ITCL_PUBLIC) {
        Itcl_ChangeCmdProtection(interp, accessCmd, ITCL_PUBLIC);
    }

    return archComp;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_DelArchComponent()
 *
 *  Destroys an Archetype component record previously created by
 *  Itk_CreateArchComponent().
 * ------------------------------------------------------------------------
 */
static void
Itk_DelArchComponent(archComp)
    ArchComponent *archComp;  /* pointer to component data */
{
    ckfree((char*)archComp);
}


/*
 * ------------------------------------------------------------------------
 *  Itk_GetArchOption()
 *
 *  Finds or creates the data representing a composite configuration
 *  option for an Archetype mega-widget.  Each option acts as a single
 *  entity, but is composed of several parts which propagate changes
 *  down to the component widgets.  If the option already exists, then
 *  the specified resource name and resource class must match the
 *  existing definition.
 *
 *  If the option is created, an initial value for is determined by
 *  querying the X11 resource database, and if this fails, the
 *  hard-wired default value is used.
 *
 *  If successful, returns TCL_OK along with a pointer to the option
 *  record.  Returns TCL_ERROR (along with an error message in the
 *  interpreter) if anything goes wrong.
 * ------------------------------------------------------------------------
 */
static int
Itk_GetArchOption(interp,info,switchName,resName,resClass,defVal,currVal,aoPtr)
    Tcl_Interp *interp;            /* interpreter managing the object */
    ArchInfo *info;                /* info for Archetype mega-widget */
    char *switchName;              /* name of command-line switch */
    char *resName;                 /* resource name in X11 database */
    char *resClass;                /* resource class name in X11 database */
    char *defVal;                  /* last-resort default value */
    char *currVal;                 /* current option value */
    ArchOption **aoPtr;            /* returns: option record */
{
    char *init = NULL;

    ArchOption *archOpt;
    Tcl_HashEntry *entry;
    int newEntry;
    char c, *ival;
    Itcl_ActiveNamespace nsToken;
    CallFrame frame;

    /*
     *  See if an option already exists with the switch name.
     *  If it does, then make sure that the given resource name
     *  and resource class match the existing definition.
     */
    entry = Tcl_CreateHashEntry(&info->options, switchName, &newEntry);
    if (!newEntry) {
        archOpt = (ArchOption*)Tcl_GetHashValue(entry);
        if (strcmp(archOpt->resName, resName) != 0) {
            Tcl_AppendResult(interp, "bad resource name ",
                "for option \"", switchName, "\": should be \"",
                archOpt->resName, "\"",
                (char*)NULL);
            return TCL_ERROR;
        }
        if (strcmp(archOpt->resClass, resClass) != 0) {
            Tcl_AppendResult(interp, "bad resource class ",
                "for option \"", switchName, "\": should be \"",
                archOpt->resClass, "\"",
                (char*)NULL);
            return TCL_ERROR;
        }
        *aoPtr = archOpt;
        return TCL_OK;
    }

    /*
     *  Otherwise, create a new option record.
     *  If this widget has a Tk window, query the X11 resource
     *  database for an initial option value.  If all else fails,
     *  use the hard-wired default value.
     */
    if (info->tkwin != NULL) {
        init = Tk_GetOption(info->tkwin, resName, resClass);
    }
    if (init == NULL) {
        init = defVal;
    }

    /*
     *  Normally, the initial value for the itk_option array is
     *  the same as the initial value for the option.  Watch
     *  out for the fixed Tk options (-class, -colormap, -screen
     *  and -visual).  Since these cannot be modified later,
     *  they must be set to their current value.
     */
    c = *(switchName+1);

    if ((c == 'c' && strcmp(switchName,"-class") == 0) ||
        (c == 'c' && strcmp(switchName,"-colormap") == 0) ||
        (c == 's' && strcmp(switchName,"-screen") == 0) ||
        (c == 'v' && strcmp(switchName,"-visual") == 0)) {
        ival = currVal;
    }
    else {
        ival = init;
    }

    /*
     *  Set the initial value in the itk_option array.
     *  Since this might be called from the itk::option-parser
     *  namespace, reinstall the object context.
     *
     *  NOTE:  A local call frame MUST be installed to gain proper
     *    access to object-specific data members.  Object-specific
     *    variables cannot be cached at the class-namespace level,
     *    and therefore must be accessed via a local call frame.
     */
    nsToken = Itcl_ActivateNamesp2(interp, info->obj->cdefn->namesp,
        (ClientData)info->obj);
    Itcl_InstallCallFrame(interp, &frame, 0, (char**)NULL);

    Tcl_SetVar2(interp, "itk_option", switchName, ival, 0);

    Itcl_RemoveCallFrame(interp);
    Itcl_DeactivateNamesp(interp, nsToken);

    /*
     *  Create the record to represent this option, and save it
     *  in the option table.
     */
    archOpt = (ArchOption*)ckalloc(sizeof(ArchOption));

    archOpt->switchName = (char*)ckalloc((unsigned)(strlen(switchName)+1));
    strcpy(archOpt->switchName, switchName);

    archOpt->resName = (char*)ckalloc((unsigned)(strlen(resName)+1));
    strcpy(archOpt->resName, resName);

    archOpt->resClass = (char*)ckalloc((unsigned)(strlen(resClass)+1));
    strcpy(archOpt->resClass, resClass);

    archOpt->init = (char*)ckalloc((unsigned)(strlen(init)+1));
    strcpy(archOpt->init, init);

    archOpt->flags = 0;
    Itcl_InitList(&archOpt->parts);

    Tcl_SetHashValue(entry, (ClientData)archOpt);
    Itk_OptListAdd(&info->order, entry);

    *aoPtr = archOpt;
    return TCL_OK;
}

/*
 * ------------------------------------------------------------------------
 *  Itk_DelArchOption()
 *
 *  Destroys an Archetype configuration option previously created by
 *  Itk_CreateArchOption().
 * ------------------------------------------------------------------------
 */
static void
Itk_DelArchOption(archOpt)
    ArchOption *archOpt;  /* pointer to option data */
{
    Itcl_ListElem *elem;
    ArchOptionPart *optPart;

    /*
     *  Delete all "parts" relating to component widgets.
     */
    elem = Itcl_FirstListElem(&archOpt->parts);
    while (elem) {
        optPart = (ArchOptionPart*)Itcl_GetListValue(elem);
        Itk_DelOptionPart(optPart);
        elem = Itcl_DeleteListElem(elem);
    }

    /*
     *  Free any remaining data.
     */
    ckfree(archOpt->switchName);
    ckfree(archOpt->resName);
    ckfree(archOpt->resClass);
    ckfree(archOpt->init);
    ckfree((char*)archOpt);
}


/*
 * ------------------------------------------------------------------------
 *  Itk_CreateOptionPart()
 *
 *  Creates the data representing a part within a configuration option
 *  for an Archetype mega-widget.  Each part has a bit of code used to
 *  apply configuration changes to some part of the mega-widget.
 *  This is characterized by a bit of ClientData, and a "config"
 *  procedure that knows how to execute it.  The ClientData is
 *  automatically disposed of by the delete proc when this option
 *  part is destroyed.
 *
 *  Option parts typically come from two sources:  Options defined
 *  in the class definition, and options propagated upward from
 *  component parts.
 *
 *  Returns a pointer to the new option part.
 * ------------------------------------------------------------------------
 */
static ArchOptionPart*
Itk_CreateOptionPart(interp,cdata,cproc,dproc,from)
    Tcl_Interp *interp;              /* interpreter handling this request */
    ClientData cdata;                /* data representing this part */
    Itk_ConfigOptionPartProc *cproc; /* proc used to apply config changes */
    Itcl_DeleteProc *dproc;          /* proc used to clean up ClientData */
    ClientData from;                 /* who contributed this option */
{
    ArchOptionPart *optPart;

    /*
     *  Create the record to represent this part of the option.
     */
    optPart = (ArchOptionPart*)ckalloc(sizeof(ArchOptionPart));
    optPart->clientData = cdata;
    optPart->configProc = cproc;
    optPart->deleteProc = dproc;
    optPart->from       = from;

    return optPart;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_AddOptionPart()
 *
 *  Integrates an option part into a composite configuration option
 *  for an Archetype mega-widget.  If a composite option does not
 *  yet exist with the specified switch name, it is created automatically.
 *
 *  Adds the option part onto the composite list, and reconfigures
 *  the widget to update this option properly.
 *
 *  Returns TCL_OK on success, or TCL_ERROR (along with an error message
 *  in the interpreter) if anything goes wrong.
 * ------------------------------------------------------------------------
 */
static int
Itk_AddOptionPart(interp,info,switchName,resName,resClass,defVal,currVal,
    optPart, raOpt)

    Tcl_Interp *interp;              /* interpreter handling this request */
    ArchInfo *info;                  /* info for Archetype mega-widget */
    char *switchName;                /* name of command-line switch */
    char *resName;                   /* resource name in X11 database */
    char *resClass;                  /* resource class name in X11 database */
    char *defVal;                    /* last-resort default value */
    char *currVal;                   /* current value (or NULL) */
    ArchOptionPart *optPart;         /* part to be added in */
    ArchOption **raOpt;              /* returns: option containing new part */
{
    int status;
    char *init;
    ArchOption *archOpt;
    Itcl_ActiveNamespace nsToken;
    CallFrame frame;

    *raOpt = NULL;

    /*
     *  Find or create a composite option for the mega-widget.
     */
    if (Itk_GetArchOption(interp, info, switchName, resName, resClass,
        defVal, currVal, &archOpt) != TCL_OK) {
        return TCL_ERROR;
    }

    /*
     *  Add the option part to the composite option.  If the
     *  composite option has already been configured, then
     *  simply update this part to the current value.  Otherwise,
     *  leave the configuration to Itk_ArchInitCmd().
     */
    Itcl_AppendList(&archOpt->parts, (ClientData)optPart);

    if ((archOpt->flags & ITK_ARCHOPT_INIT) != 0) {
        /*
         *  NOTE:  A local call frame MUST be installed to gain proper
         *    access to object-specific data members.  Object-specific
         *    variables cannot be cached at the class-namespace level,
         *    and therefore must be accessed via a local call frame.
         */
        nsToken = Itcl_ActivateNamesp2(interp, info->obj->cdefn->namesp,
            (ClientData)info->obj);
        Itcl_InstallCallFrame(interp, &frame, 0, (char**)NULL);

        init = Tcl_GetVar2(interp, "itk_option", archOpt->switchName, 0);

        Itcl_RemoveCallFrame(interp);
        Itcl_DeactivateNamesp(interp, nsToken);

        if (!init) {
            Itk_ArchOptAccessError(interp, info, archOpt);
            return TCL_ERROR;
        }

        if (!currVal || (strcmp(init,currVal) != 0)) {
            status  = (*optPart->configProc)(interp, info->obj,
                optPart->clientData, init);

            if (status != TCL_OK) {
                Itk_ArchOptConfigError(interp, info, archOpt);
                return TCL_ERROR;
            }
        }
    }

    *raOpt = archOpt;
    return TCL_OK;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_FindArchOptionPart()
 *
 *  Searches for a specific piece of a composite configuration option
 *  for an Archetype mega-widget.  The specified name is treated as the
 *  "switch" name (e.g., "-option"), but this procedure will recognize
 *  it even without the leading "-".
 *
 *  Returns a pointer to the option with the matching switch name and
 *  source, or NULL if the option is not recognized.
 * ------------------------------------------------------------------------
 */
static ArchOptionPart*
Itk_FindArchOptionPart(info,switchName,from)
    ArchInfo *info;                /* info for Archetype mega-widget */
    char *switchName;              /* name of command-line switch */
    ClientData from;               /* who contributed this option */
{
    ArchOptionPart *optPart = NULL;

    Tcl_DString buffer;
    Tcl_HashEntry *entry;
    ArchOption *archOpt;
    ArchOptionPart *op;
    Itcl_ListElem *elem;

    /*
     *  If the switch does not have a leading "-", add it on.
     */
    Tcl_DStringInit(&buffer);
    if (*switchName != '-') {
        Tcl_DStringAppend(&buffer, "-", -1);
        Tcl_DStringAppend(&buffer, switchName, -1);
        switchName = Tcl_DStringValue(&buffer);
    }

    /*
     *  Look for a composite option, and then for a part with the
     *  matching source.
     */
    entry = Tcl_FindHashEntry(&info->options, switchName);

    if (entry) {
        archOpt = (ArchOption*)Tcl_GetHashValue(entry);
        elem = Itcl_FirstListElem(&archOpt->parts);
        while (elem) {
            op = (ArchOptionPart*)Itcl_GetListValue(elem);
            if (op->from == from) {
                optPart = op;
                break;
            }
            elem = Itcl_NextListElem(elem);
        }
    }

    Tcl_DStringFree(&buffer);
    return optPart;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_RemoveArchOptionPart()
 *
 *  Searches for a specific piece of a composite configuration option
 *  for an Archetype mega-widget.  The specified name is treated as the
 *  "switch" name (e.g., "-option"), but this procedure will recognize
 *  it even without the leading "-".  If an option part with the
 *  specified name and source is found on the list, it is removed.
 *
 *  NOTE:  This procedure assumes that there is a valid object context
 *    and a call frame supporting object data member access.  It is
 *    usually called from within the methods of the Archetype base
 *    class, so this is a good assumption.  If it is called anywhere
 *    else, the caller is responsible for installing the object context
 *    and setting up a call frame.
 *
 *  Returns non-zero if the part was found and removed, and 0 otherwise.
 * ------------------------------------------------------------------------
 */
static int
Itk_RemoveArchOptionPart(info,switchName,from)
    ArchInfo *info;                /* info for Archetype mega-widget */
    char *switchName;              /* name of command-line switch */
    ClientData from;               /* who contributed this option */
{
    int status = 0;

    Tcl_DString buffer;
    Tcl_HashEntry *entry;
    ArchOption *archOpt;
    ArchOptionPart *op;
    Itcl_ListElem *elem;

    /*
     *  If the switch does not have a leading "-", add it on.
     */
    Tcl_DStringInit(&buffer);
    if (*switchName != '-') {
        Tcl_DStringAppend(&buffer, "-", -1);
        Tcl_DStringAppend(&buffer, switchName, -1);
        switchName = Tcl_DStringValue(&buffer);
    }

    /*
     *  Look for a composite option, and then for a part with the
     *  matching source.  If found, remove it.
     */
    entry = Tcl_FindHashEntry(&info->options, switchName);

    if (entry) {
        archOpt = (ArchOption*)Tcl_GetHashValue(entry);
        elem = Itcl_FirstListElem(&archOpt->parts);
        while (elem) {
            op = (ArchOptionPart*)Itcl_GetListValue(elem);
            if (op->from == from) {
                Itk_DelOptionPart(op);
                status = 1;
                elem = Itcl_DeleteListElem(elem);
            }
            else {
                elem = Itcl_NextListElem(elem);
            }
        }

        /*
         *  If this option is now dead (no parts left), then
         *  remove it from the widget.  Be careful to delete it
         *  from the "itk_option" array as well.
         */
        if (Itcl_GetListLength(&archOpt->parts) == 0) {
            Tcl_UnsetVar2(info->obj->cdefn->interp,
                "itk_option", archOpt->switchName, 0);

            Itk_DelArchOption(archOpt);
            Itk_OptListRemove(&info->order, entry);
            Tcl_DeleteHashEntry(entry);
        }
    }

    Tcl_DStringFree(&buffer);
    return status;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_IgnoreArchOptionPart()
 *
 *  Removes the specified part from a composite configuration option
 *  for an Archetype mega-widget.  This is usually called before
 *  keeping or renaming an option, to make sure that the option
 *  is not already integrated elsewhere on the composite list.
 *  This also handles the action of "ignoring" a configuration option.
 *
 *  NOTE:  This procedure assumes that there is a valid object context
 *    and a call frame supporting object data member access.  It is
 *    usually called from within the methods of the Archetype base
 *    class, so this is a good assumption.  If it is called anywhere
 *    else, the caller is responsible for installing the object context
 *    and setting up a call frame.
 *
 *  Returns non-zero if the part was found and removed, and 0 otherwise.
 * ------------------------------------------------------------------------
 */
static int
Itk_IgnoreArchOptionPart(info,opt)
    ArchInfo *info;                /* info for Archetype mega-widget */
    GenericConfigOpt *opt;         /* part to be ignored */
{
    int status = 0;

    Tcl_HashEntry *entry;
    ArchOptionPart *op;
    Itcl_ListElem *elem;

    /*
     *  If the part is not integrated, then do nothing.
     *  Otherwise, find the missing part and remove it.
     */
    if (opt->integrated) {
        elem = Itcl_FirstListElem(&opt->integrated->parts);
        while (elem) {
            op = (ArchOptionPart*)Itcl_GetListValue(elem);
            if (op == opt->optPart) {
                Itk_DelOptionPart(op);
                status = 1;
                elem = Itcl_DeleteListElem(elem);
            }
            else {
                elem = Itcl_NextListElem(elem);
            }
        }

        /*
         *  If this option is now dead (no parts left), then
         *  remove it from the widget.  Be careful to delete it
         *  from the "itk_option" array as well.
         */
        if (Itcl_GetListLength(&opt->integrated->parts) == 0) {
            Tcl_UnsetVar2(info->obj->cdefn->interp,
                "itk_option", opt->integrated->switchName, 0);

            entry = Tcl_FindHashEntry(&info->options,
                opt->integrated->switchName);

            if (entry) {
                Itk_OptListRemove(&info->order, entry);
                Tcl_DeleteHashEntry(entry);
            }
            Itk_DelArchOption(opt->integrated);
        }

        /*
         *  Forget that this part was ever integrated.
         */
        opt->integrated = NULL;
        opt->optPart = NULL;
    }
    return status;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_DelOptionPart()
 *
 *  Destroys part of an Archetype configuration option created by
 *  Itk_CreateOptionPart().
 * ------------------------------------------------------------------------
 */
static void
Itk_DelOptionPart(optPart)
    ArchOptionPart *optPart;  /* option part data to be destroyed */
{
    if (optPart->clientData && optPart->deleteProc) {
        (*optPart->deleteProc)(optPart->clientData);
    }
    ckfree((char*)optPart);
}


/*
 * ------------------------------------------------------------------------
 *  Itk_CreateGenericOptTable()
 *
 *  Parses a string describing a widget's configuration options (of the
 *  form returned by the usual widget "configure" method) and creates
 *  a hash table for easy lookup of option information.  Entries in
 *  the hash table are indexed by switch names like "-background".
 *  Values are GenericConfigOpt records.  Alias options like "-bg" are
 *  ignored.
 *
 *  This table is used by option parsing commands in "itk::option-parser"
 *  to validate widget options.
 *
 *  Returns a pointer to a new hash table, which should later be freed
 *  via Itk_DelGenericOptTable().  Returns NULL if an error is found in
 *  the configuration list.
 * ------------------------------------------------------------------------
 */
static Tcl_HashTable*
Itk_CreateGenericOptTable(interp,options)
    Tcl_Interp *interp;          /* interpreter handling this request */
    char *options;               /* string description of config options */
{
    int confc;
    char **confv = NULL;
    int optc;
    char **optv = NULL;

    int i, newEntry;
    Tcl_HashTable *tPtr;
    Tcl_HashEntry *entry;
    GenericConfigOpt *info;

    tPtr = (Tcl_HashTable*)ckalloc(sizeof(Tcl_HashTable));
    Tcl_InitHashTable(tPtr, TCL_STRING_KEYS);

    /*
     *  Split the list of options and store each one in the table.
     *  Only consider options with all 5 required components.  Avoid
     *  aliases like "-bg".
     */
    if (Tcl_SplitList(interp, options, &confc,&confv) != TCL_OK) {
        goto tableFail;
    }
    for (i=0; i < confc; i++) {
        if (Tcl_SplitList(interp, confv[i], &optc,&optv) != TCL_OK) {
            goto tableFail;
        }
        if (optc == 5) {    /* avoid aliased options */
            entry = Tcl_CreateHashEntry(tPtr, optv[0], &newEntry);
            if (newEntry) {
                info = (GenericConfigOpt*)ckalloc(sizeof(GenericConfigOpt));
                info->switchName = optv[0];
                info->resName    = optv[1];
                info->resClass   = optv[2];
                info->init       = optv[3];
                info->value      = optv[4];
                info->storage    = optv;
                info->integrated = NULL;
                info->optPart    = NULL;
                Tcl_SetHashValue(entry, (ClientData)info);
            }
        }
        else {
            ckfree((char*)optv);
        }
    }

    ckfree((char*)confv);
    return tPtr;

tableFail:
    if (confv) {
        ckfree((char*)confv);
    }
    Itk_DelGenericOptTable(tPtr);
    return NULL;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_DelGenericOptTable()
 *
 *  Destroys an option table previously created by
 *  Itk_CreateGenericOptTable() and frees all memory associated with it.
 *  Should be called whenever a table is no longer needed, to free up
 *  resources.
 * ------------------------------------------------------------------------
 */
static void
Itk_DelGenericOptTable(tPtr)
    Tcl_HashTable *tPtr;  /* option table to be destroyed */
{
    Tcl_HashEntry *entry;
    Tcl_HashSearch place;
    GenericConfigOpt *info;

    /*
     *  Scan through all options in the table and free entries.
     */
    entry = Tcl_FirstHashEntry(tPtr, &place);
    while (entry) {
        info = (GenericConfigOpt*)Tcl_GetHashValue(entry);
        ckfree((char*)info->storage);
        ckfree((char*)info);
        entry = Tcl_NextHashEntry(&place);
    }

    Tcl_DeleteHashTable(tPtr);
    ckfree((char*)tPtr);
}


/*
 * ------------------------------------------------------------------------
 *  Itk_CreateGenericOpt()
 *
 *  Parses a string describing a widget's configuration option (of the
 *  form returned by the usual widget "configure" method) and creates
 *  a representation for one option.  Similar to
 *  Itk_CreateGenericOptTable(), but only handles one option at a
 *  time.
 *
 *  Returns a pointer to the option info, which should later be freed
 *  via Itk_DelGenericOpt().  Returns NULL (along with an error
 *  message in the interpreter) if an error is found.
 *
 *  SIDE EFFECT:  Resets the interpreter result.
 * ------------------------------------------------------------------------
 */
static GenericConfigOpt*
Itk_CreateGenericOpt(interp,switchName,accessCmd)
    Tcl_Interp *interp;          /* interpreter handling this request */
    char *switchName;            /* command-line switch for option */
    Tcl_Command accessCmd;       /* access command for component */
{
    GenericConfigOpt *genericOpt = NULL;

    int optc;
    char **optv;
    char *info;
    Tcl_DString buffer, code;

    /*
     *  If the switch does not have a leading "-", add it on.
     */
    Tcl_DStringInit(&buffer);
    if (*switchName != '-') {
        Tcl_DStringAppend(&buffer, "-", -1);
        Tcl_DStringAppend(&buffer, switchName, -1);
        switchName = Tcl_DStringValue(&buffer);
    }

    /*
     *  Build a "configure" command to query info for the requested
     *  option.  Evaluate the command and get option info.
     */
    Tcl_DStringInit(&code);
    Tcl_DStringAppendElement(&code, Tcl_GetCommandPath(interp, accessCmd));
    Tcl_DStringAppendElement(&code, "configure");
    Tcl_DStringAppendElement(&code, switchName);

    if (Tcl_Eval(interp, Tcl_DStringValue(&code)) != TCL_OK) {
        goto optionDone;
    }

    /*
     *  Only consider options with all 5 required components.  Avoid
     *  aliases like "-bg".
     */
    Tcl_DStringTrunc(&buffer, 0);
    Tcl_DStringAppend(&buffer, interp->result, -1);
    info = Tcl_DStringValue(&buffer);

    if (Tcl_SplitList(interp, info, &optc,&optv) != TCL_OK) {
        goto optionDone;
    }
    if (optc == 5) {    /* avoid aliased options */
        genericOpt = (GenericConfigOpt*)ckalloc(sizeof(GenericConfigOpt));
        genericOpt->switchName = optv[0];
        genericOpt->resName    = optv[1];
        genericOpt->resClass   = optv[2];
        genericOpt->init       = optv[3];
        genericOpt->value      = optv[4];
        genericOpt->storage    = optv;
        genericOpt->integrated = NULL;
        genericOpt->optPart    = NULL;
    }
    else {
        ckfree((char*)optv);
    }

optionDone:
    Tcl_DStringFree(&buffer);
    Tcl_DStringFree(&code);

    if (genericOpt) {
        Tcl_ResetResult(interp);
    }
    return genericOpt;
}


/*
 * ------------------------------------------------------------------------
 *  Itk_DelGenericOpt()
 *
 *  Destroys a generic option previously created by Itk_CreateGenericOpt()
 *  and frees all memory associated with it.  Should be called whenever
 *  an option representation is no longer needed, to free up resources.
 * ------------------------------------------------------------------------
 */
static void
Itk_DelGenericOpt(opt)
    GenericConfigOpt *opt;  /* option info to be destroyed */
{
    ckfree((char*)opt->storage);
    ckfree((char*)opt);
}
